from typing import NewType, TypeVar

import string
from typing import Union
from textual import events
from textual.app import App
from textual.keys import Keys
from textual.widgets import Placeholder, ScrollView, Footer, TreeControl, TreeClick, TreeNode
from textual.widget import Widget
from textual.reactive import Reactive
from textual.views import DockView
from rich.align import Align
from rich import box
from rich.console import RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree
from rich.markdown import Markdown
from rich.style import Style
import rich.repr
import numpy as np
import copy

import bluesky as bs
from bluesky.network import context as ctx
from bluesky.network.client import Client
from bluesky.tools.misc import tim2txt
from bluesky.tools.aero import ft, kts, nm, fpm

NodeID = NewType("NodeID", int)
NodeDataType = TypeVar("NodeDataType")
class ConsoleClient(Client):
    '''
        Subclassed Client with a timer to periodically check for incoming data,
        an overridden event function to handle data, and a stack function to
        send stack commands to BlueSky.
    '''
    modes = ['Init', 'Hold', 'Operate', 'End']

    def __init__(self, actnode_topics=b''):
        super().__init__(actnode_topics)
        self.subscribe(b'SIMINFO').connect(self.on_siminfo_received)
        self.subscribe(b'ACDATA').connect(self.on_acdata_received)
        
        self.count = 0
        self.nodes = dict()

    def event(self, name, data, sender_id):
        ''' Overridden event function. '''
        pass

    def echo(self, text, flags=None, sender_id=None):
        ''' Overload Client's echo function. '''
        if ConsoleUI.instance is not None:
            ConsoleUI.instance.echo(text, flags)

    def actnode_changed(self, newact):
        pass

    def on_siminfo_received(self, data):
        if ConsoleUI.instance is not None:
            speed, simdt, simt, simutc, ntraf, state, scenname = data
            simt = tim2txt(simt)[:-3]
            self.setNodeInfo(ctx.sender_id, simt, scenname)
            if ctx.sender_id == bs.net.act_id:
                ConsoleUI.instance.set_infoline(f'[b]t:[/b] {simt} [b]dt:[/b] {simdt} [b]Speed:[/b] {speed:.1f} [b]UTC:[/b] {simutc} [b]Mode:[/b] {self.modes[state]} [b]Aircraft:[/b] {ntraf}')

            # concate times of all nodes
            node_times = "".join([self.nodes[key]['time'] for key in self.nodes])

            # send to tui
            ConsoleUI.instance.set_nodes(copy.deepcopy(self.nodes), node_times)

    def on_acdata_received(self, data):
        self.extend_node_data(data, ctx.sender_id)
        if ctx.sender_id == bs.net.act_id:
            gen_data, table_data = self.get_traffic(data)
            ConsoleUI.instance.set_traffic(gen_data, table_data, ctx.sender_id)

    def extend_node_data(self, data, connid):
        # check if it is inside self.nodes
        if connid in self.nodes:
            self.nodes[connid]['nair'] = len(data['id'])
            self.nodes[connid]['nconf_cur'] = data['nconf_cur']
            self.nodes[connid]['nconf_tot'] = data['nconf_tot']
            self.nodes[connid]['nlos_cur'] = data['nlos_cur']
            self.nodes[connid]['nlos_tot'] = data['nlos_tot']

    def get_traffic(self, data):
        
        # general data
        gen_data = dict()
        gen_data['simt'] = data['simt'] 
        
        # only keep some info for table
        table_data = dict()

        table_data['id'] = data['id']
        table_data['lat'] = np.array(np.round(data['lat'],4), dtype=str)
        table_data['lon'] = np.array(np.round(data['lon'],4), dtype=str)
        table_data['alt (ft)'] = np.array(np.round(data['alt']/ft, 0), dtype=str)
        table_data['tas (kts)'] = np.array(np.round(data['tas']/kts,2), dtype=str)
        table_data['cas (kts)'] = np.array(np.round(data['cas']/kts,2), dtype=str)
        table_data['inconf'] = np.array(np.array(data['inconf'], dtype=bool), dtype=str)
        table_data['tcpamax (s)'] = np.array(np.round(data['tcpamax'],2), dtype=str)
        table_data['rpz (nm)'] = np.array(np.round(data['rpz']/nm,2), dtype=str)
        table_data['vs (fpm)'] = np.array(np.round(data['vs']/fpm,2), dtype=str)
        table_data['vmin (kts)'] = np.array(np.round(data['vmin']/kts,2), dtype=str)
        table_data['vmax (kts)'] = np.array(np.round(data['vmax']/kts,2), dtype=str)

        return gen_data, table_data
        
    def setNodeInfo(self, connid, time, scenname):
        
        # check if it is inside self.nodes
        if connid in self.nodes:
            self.nodes[connid]['scenename'] = scenname
            self.nodes[connid]['time'] = time
        
        else:
            node_count = self.count + 1
            self.nodes[connid] = {'num': f'{node_count}','scenename': scenname, 
                                'time': time, 'nair': 0, 'nconf_cur': 0, 'nconf_tot': 0,
                                'nlos_cur': 0, 'nlos_tot': 0}
            self.count += 1
            
class Echobox(ScrollView):
    text = Reactive("")

    async def watch_text(self, text):
        await self.update(Panel(Text(text), height=max(8, 2 + text.count('\n')), box=box.SIMPLE, style=Style(bgcolor="grey53")))

    def set_text(self, text: str):
        self.text = text

class Textline(Widget):
    text = Reactive("")
    style = Style(bgcolor="grey37")

    def __init__(self, text="", name=None):
        super().__init__(name)
        self.text = text

    def render(self) -> RenderableType:
        return self.text

    def set_text(self, text):
        self.text = text

class NodeInfo(Widget):
    table: Union[Reactive[Table], Table] = Reactive(Table())

    def __init__(self, name=None):
        super().__init__(name)
        self.table = Table()

        self.regular_style = Style(color='bright_green', bgcolor="bright_black")
        self.actnode_style = Style(color='bright_yellow', bgcolor="black")

    def render(self) -> RenderableType:
        return self.table

    def set_nodes(self, nodedict: dict):
        # add rows to table
        self.table = Table(title='[bold blue]Nodes', box=box.MINIMAL_DOUBLE_HEAD)
        self.table.add_column('Node #', header_style=Style(color="magenta"))
        self.table.add_column('Scenario', header_style=Style(color="magenta"))
        self.table.add_column('Time', header_style=Style(color="magenta"))
        self.table.add_column('Aircraft', header_style=Style(color="magenta"))
        self.table.add_column('Current conflicts', header_style=Style(color="magenta"))
        self.table.add_column('Total conflicts', header_style=Style(color="magenta"))
        self.table.add_column('Current LOS', header_style=Style(color="magenta"))
        self.table.add_column('Total LOS', header_style=Style(color="magenta"))

        for node_id, node in nodedict.items():
            
            if bs.net.act == node_id:
               self.table.add_row(node['num'], node['scenename'], node['time'],
                                    f'{node["nair"]}',
                                    f'{node["nconf_cur"]}', f'{node["nconf_tot"]}', 
                                    f'{node["nlos_cur"]}', f'{node["nlos_tot"]}', 
                                style=self.actnode_style)
            else:
                self.table.add_row(node['num'], node['scenename'], node['time'],
                                    f'{node["nair"]}',
                                    f'{node["nconf_cur"]}', f'{node["nconf_tot"]}', 
                                    f'{node["nlos_cur"]}', f'{node["nlos_tot"]}',
                                style=self.regular_style)

class Traffic(Widget):
    table: Union[Reactive[Table], Table] = Reactive(Table())

    def __init__(self, name=None):
        super().__init__(name)
        self.table = Table()

    def render(self) -> RenderableType:
        return self.table

    def set_traffic(self, trafict: dict, active_node_num: float):
        # add rows to table
        self.table = Table(title=f'[bold blue]Traffic for node # {active_node_num}', box=box.MINIMAL_DOUBLE_HEAD)

        # add the columns
        for col in trafict.keys():
            self.table.add_column(col, header_style=Style(color="magenta"))
        
        ntraf = len(trafict['id'])

        if ntraf > 0:
            for i in range(ntraf):
                # limit traffic to 100 aircraft on screen
                if i > 100:
                    break
                row = []
                for col in trafict.keys():
                    table_value = trafict[col][i]
                    row.append(table_value)
                self.table.add_row(*row, style=Style(color='bright_green' ,bgcolor="bright_black"))

class NodeTree(TreeControl):
    nodenums = Reactive("")
    nnodes = Reactive(0)

    def __init__(self, label=None, *args):
        super().__init__(label, *args)
        self.nnodes = 0
        self.original_label = label
        self.root.tree.guide_style = "green4"
        self._tree.style = Style(color="deep_sky_blue1")

    def __reinit__(self):
        # reinitialize tree for new nodes
        self._tree.label = self.root
        self.data = {}
        self.id = NodeID(0)
        self.nodes: dict[NodeID, TreeNode[NodeDataType]] = {}
        self._tree = Tree(self.original_label)
        self.root: TreeNode[NodeDataType] = TreeNode(
            None, self.id, self, self._tree, self.original_label, {}
        )
        self._tree.label = self.root
        self.nodes[NodeID(self.id)] = self.root
        self.root.tree.guide_style = "blue"
        self._tree.style = Style(color="deep_sky_blue1")

    async def handle_tree_click(self, message: TreeClick[dict]):
        """Called in response to a tree click."""
        new_active_node = message.node.data.get("node_id")
        # convert active node string to bytes
        bs.net.actnode(new_active_node)

    async def watch_nnodes(self, nnodes):

        # reinitialize if nodes exist
        if list(self.nodes.keys())[1:]:
            self.__reinit__()

        # reload nodes
        for node in self.nodenums:
            node_id = self.node_ids[self.nodenums.index(node)]
            await self.add(self.root.id, node, {"node_num": node, "node_id": node_id})
        await self.root.expand()

    def set_nodedict(self, nodedict):

        self.nodenums = []
        self.node_ids = []
        for node_id, node in nodedict.items():
            nodenum, node_id = node['num'], node_id
            self.nodenums.append(f'Node {nodenum}')
            self.node_ids.append(node_id)
        self.nnodes = len(nodedict)


class ConsoleUI(App):
    cmdtext: Union[Reactive[str], str] = Reactive("")
    echotext: Union[Reactive[str], str] = Reactive("")
    infotext: Union[Reactive[str], str] = Reactive("")
    nodedict = Reactive(dict())
    nodetimes = Reactive("0")
    trafsimt = Reactive("")
    ntraf = Reactive(0)
    nnodes = Reactive(0)
    active_node_num = Reactive(0)

    cmdbox: Textline
    echobox: Echobox
    infoline: Textline
    nodeinfo: NodeInfo
    traffic: Traffic
    tree: NodeTree
    instance: App
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        ConsoleUI.instance = self
        
    def echo(self, text, flags=None):
        self.echotext = text + '\n' + self.echotext

    def set_infoline(self, text):
        self.infotext = text
        
    def set_nodes(self, nodes, nodetimes):
        self.nodedict = nodes
        self.nnodes = len(nodes)
        self.nodetimes = nodetimes

    def set_traffic(self, gen_data, traffic, active_node):
        self.trafdict = traffic
        self.trafsimt = str(gen_data['simt'])
        self.ntraf = len(traffic['id'])
        self.active_node_num = self.nodedict[active_node]['num'] if self.nodedict else 1
        
    async def on_key(self, key: events.Key):
        if key.key == Keys.ControlH:
            self.cmdtext = self.cmdtext[:-1]
        elif key.key == Keys.Delete:
            self.cmdtext = ""
        elif key.key == Keys.Enter:
            self.echotext = self.cmdtext + '\n' + self.echotext
            bs.stack.stack(self.cmdtext)
            self.cmdtext = ""
        elif key.key == Keys.Escape:
            await self.action_quit()
        elif key.key in string.printable:
            self.cmdtext += key.key

    async def watch_infotext(self, infotext):
        self.infoline.set_text(f"[black]Current node:[/black] {infotext}")

    async def watch_cmdtext(self, cmdtext):
        self.cmdbox.set_text(f"[blue]>>[/blue] {cmdtext}")
        
    async def watch_nodetimes(self, nodetimes):
        self.nodeinfo.set_nodes(self.nodedict)
        self.tree.set_nodedict(self.nodedict)
    
    async def watch_nnodes(self, nnodes):
        await self.nodebody.update(self.nodeinfo)

    async def watch_echotext(self, echotext):
        self.echobox.set_text(echotext)

    async def watch_ntraf(self, ntraf):
        await self.trafbody.update(self.traffic)

    async def watch_trafsimt(self, trafsimt):
        self.traffic.set_traffic(self.trafdict, self.active_node_num)
    
    async def watch_active_node_num(self, active_node_num):
        self.nodeinfo.set_nodes(self.nodedict)

    async def on_mount(self, event: events.Mount):
        self.cmdbox = Textline("[blue]>>[/blue]")
        self.echobox = Echobox(Panel(Text(), height=8, box=box.SIMPLE, style=Style(bgcolor="grey53")))
        self.infoline = Textline("[black]Current node: [/black]")
        self.nodeinfo = NodeInfo(name="nodeinfo")
        self.traffic = Traffic(name="traffic")
        self.tree = NodeTree("Switch Nodes", {})

        await self.bind(Keys.Escape, "quit", "Quit")
        await self.bind(Keys.ControlT, "view.toggle('trafficbody')", "Show traffic")
        await self.bind(Keys.ControlB, "view.toggle('nodedock')", "Show batch")

        await self.view.dock(Footer(), edge="bottom", size=1)
        await self.view.dock(self.cmdbox, edge="bottom", size=1)

        echorow = DockView()
        self.treebody = ScrollView(self.tree)
        await echorow.dock(self.treebody, edge="right", size=20)
        await echorow.dock(self.echobox, edge="left")
        
        await self.view.dock(echorow, edge="bottom", size=8)
        await self.view.dock(self.infoline, edge="bottom", size=1)
        
        self.trafbody = ScrollView(self.traffic, name="trafficbody")
        await self.view.dock(self.trafbody, edge='top')

        self.nodebody = ScrollView(self.nodeinfo, name="nodeinfobody")
        await self.view.dock(self.nodebody, edge='top')

        await self.set_focus(self.cmdbox)

        self.set_interval(0.2, bs.net.update, name='Network')

def start(hostname=None):
    ''' Create and start BlueSky text-based client. '''
    bsclient = ConsoleClient()
    bsclient.connect(hostname=hostname)

    ConsoleUI.run(log="textual.log")


from pathlib import Path
import cdsapi
import datetime
import numpy as np
import bluesky as bs
import netCDF4 as nc
from bluesky import stack
from bluesky.core import timed_function
from bluesky.traffic.windsim import WindSim


datadir = Path('')


def init_plugin():
    global datadir
    datadir = bs.resource(bs.settings.data_path) / 'NetCDF'

    if not datadir.is_dir():
        datadir.mkdir()

    global windecmwf
    windecmwf = WindECMWF()

    config = {
        'plugin_name': 'WINDECMWF',
        'plugin_type': 'sim'
    }

    return config

class WindECMWF(WindSim):
    def __init__(self):
        super().__init__()
        self.year  = 0
        self.month = 0
        self.day   = 0
        self.hour  = 0
        self.lat0  = -90
        self.lon0  = -180
        self.lat1  = 90
        self.lon1  = 180

        # Switch for periodic loading of new GFS data
        self.autoload = True
        
    def fetch_nc(self, year, month, day):
        """
        Retrieve weather data via the CDS API for multiple pressure levels
        """
        
        ymd = "%04d%02d%02d" % (year, month, day)
        
        fname = f'p_levels_{ymd}.nc'
        fpath = datadir / fname
        
        if not fpath.is_file():
            stack.echo("Downloading file, please wait...")
    
            # Set client
            c = cdsapi.Client()
            
            # Retrieve data 
            c.retrieve(
                'reanalysis-era5-pressure-levels',
                {
                    'product_type': 'reanalysis',
                    'format': 'netcdf',
                    'pressure_level': [
                        '100', '125', '150', 
                        '175', '200', '225',
                        '250', '300', '350',
                        '400', '450', '500',
                        '550', '600', '650', 
                        '700', '750', '775',
                        '800'
                    ],
                    'year': year,
                    'month': month,
                    'day': day,
                    'time': [
                        '00:00', '03:00', '06:00',
                        '09:00', '12:00', '15:00',
                        '18:00', '21:00',
                    ],
                    'variable': [
                        'u_component_of_wind', 'v_component_of_wind'
                    ],
                },
                fpath)
    
        stack.echo("Download completed.")
        netcdf = nc.Dataset(fpath, mode='r')
    
        return netcdf


    def extract_wind(self, netcdf, lat0, lon0, lat1, lon1, hour):

        # Load reanalysis data 
        level = netcdf['level'][:].data
        lats  = netcdf['latitude'][:].data
        lons  = netcdf['longitude'][:].data
        vxs_  = netcdf['u'][:].squeeze().data
        vys_  = netcdf['v'][:].squeeze().data
        
        # Close data for performance
        netcdf.close()   
        
        # Transform pressure levels to altitude
        p = level * 100
        h = (1 - (p / 101325.0)**0.190264) * 44330.76923    # in meters
        
        # Set hour to rounded hour
        hour = round(hour/3)
        
        # Construct 2D array of all data points
        lats_ = np.tile(np.repeat(lats, len(lons)), len(level))
        lons_ = np.tile(lons, len(lats)*len(level))
        alts_ = np.repeat(h, len(lats)*len(lons))       
        vxs_  = vxs_[hour,:,:,:].flatten()
        vys_  = vys_[hour,:,:,:].flatten()
            
        # Convert longitudes
        lons_ = (lons_ + 180) % 360.0 - 180.0     # convert range from 0~360 to -180~180
        
        # Reduce area based on lat lon limits
        lat0_ = min(lat0, lat1)
        lat1_ = max(lat0, lat1)
        lon0_ = min(lon0, lon1)
        lon1_ = max(lon0, lon1)

        mask = (lats_ >= lat0_) & (lats_ <= lat1_) & (lons_ >= lon0_) & (lons_ <= lon1_)

        data = np.array([lats_[mask], lons_[mask], alts_[mask], vxs_[mask], vys_[mask]])

        return data

    @stack.command(name='WINDECMWF')
    def loadwind(self, lat0: 'lat', lon0: 'lon', lat1: 'lat', lon1: 'lon',
               year: int=None, month: int=None, day: int=None, hour: int=None):
        ''' WINDECMWF: Load a windfield directly from NOAA database.

            Arguments:
            - lat0, lon0, lat1, lon1 [deg]: Bounding box in which to generate wind field
            - year, month, day, hour: Date and time of wind data (optional, will use
              current simulation UTC if not specified).
        '''
        self.lat0, self.lon0, self.lat1, self.lon1 =  min(lat0, lat1), \
                              min(lon0, lon1), max(lat0, lat1), max(lon0, lon1)
        self.year = year or bs.sim.utc.year
        self.month = month or bs.sim.utc.month
        self.day = day or bs.sim.utc.day
        self.hour = hour or bs.sim.utc.hour

        # round hour to 3 hours
        self.hour  = round(self.hour/3) * 3
        if self.hour == 24:
            ymd0 = "%04d%02d%02d" % (self.year, self.month, self.day)
            ymd1 = (datetime.datetime.strptime(ymd0, '%Y%m%d') + 
                    datetime.timedelta(days=1))
            self.year  = ymd1.year
            self.month = ymd1.month
            self.day   = ymd1.day
            self.hour  = 0

        txt = "Loading wind field for %s-%s-%s..." % (self.year, self.month, self.day)
        stack.echo("%s" % txt)

        netcdf = self.fetch_nc(self.year, self.month, self.day)

        if netcdf is None or self.lat0 == self.lat1 or self.lon0 == self.lon1:
            return False, "Wind data non-existend in area [%d, %d], [%d, %d]. " \
                % (self.lat0, self.lat1, self.lon0, self.lon1) \
                + "time: %04d-%02d-%02d" \
                % (self.year, self.month, self.day)

        # first clear exisiting wind field
        self.clear()

        # add new wind field
        data = self.extract_wind(netcdf, self.lat0, self.lon0, self.lat1, self.lon1, self.hour).T

        data = data[np.lexsort((data[:, 2], data[:, 1], data[:, 0]))] # Sort by lat, lon, alt
        reshapefactor = int((1 + (max(self.lat0, self.lat1) - min(self.lat0, self.lat1))*4) * \
                            (1 + (max(self.lon0, self.lon1) - min(self.lon0, self.lon1))*4))

        lat     = np.reshape(data[:,0], (reshapefactor, -1)).T[0,:]
        lon     = np.reshape(data[:,1], (reshapefactor, -1)).T[0,:]
        veast   = np.reshape(data[:,3], (reshapefactor, -1)).T
        vnorth  = np.reshape(data[:,4], (reshapefactor, -1)).T
        windalt = np.reshape(data[:,2], (reshapefactor, -1)).T[:,0]

        self.addpointvne(lat, lon, vnorth, veast, windalt)        

        return True, "Wind field updated in area [%d, %d], [%d, %d]. " \
            % (self.lat0, self.lat1, self.lon0, self.lon1) \
            + "time: %04d-%02d-%02d" \
            % (self.year, self.month, self.day)

    @timed_function(name='WINDECMWF', dt=3600)
    def update(self):
        if self.autoload:
            _, txt = self.loadwind(self.lat0, self.lon0, self.lat1, self.lon1)
            stack.echo("%s" % txt)


''' BlueSky base stack commands. '''
import webbrowser
from pathlib import Path

import bluesky as bs
from bluesky import settings
from bluesky.core import select_implementation, simtime, varexplorer as ve
from bluesky.tools import geo, aero, areafilter, plotter
from bluesky.tools.calculator import calculator
from bluesky.stack.cmdparser import append_commands


def initbasecmds():
    ''' Initialise BlueSky base stack commands. '''
    # Command dictionary with command as key, gives a list with:
    #
    #         command: [ helptext ,
    #                    arglist ,
    #                    function to call,
    #                    description in one line ]
    #
    # Regarding the arglist:
    #    - Separate aruments with a comma ","
    #    - Enclose optional arguments with "[" and "]"
    #    - Separate different argument type variants in one argument with "/"
    #    - Repeat last one using "..." ,    (see e.g. WIND or POLY)
    #
    # Argtypes = syntax parsing (see below in this module for parsing):
    #
    #   acid      = aircraft id (text => index)
    #   alt       = altitude (FL250, 25000  ft+. meters)
    #   spd       = CAS or Mach (when <1)   => m/s
    #   hdg       = heading in degrees, True or Magnetic
    #
    #   float     = plain float
    #   int       = integer
    #   txt       = text will be converted to upper case
    #               (for keywords, navaids, flags, waypoints,acid etc)
    #   word      = single, case sensitive word
    #   string    = case sensitive string
    #   on/off    = text => boolean
    #
    #   latlon    = converts acid, wpt, airport etc => lat,lon (deg) so 2 args!
    #   wpt       = converts postext or lat,lon into a text string,
    #               to be used as named waypoint
    #   wpinroute = text string with name of waypoint in route
    #   pandir    = text with LEFT, RIGHT, UP/ABOVE or DOWN
    #
    # Below this dictionary also a dictionary of synonym commands is given
    #
    # --------------------------------------------------------------------
    cmddict = {
        "ADDNODES": [
            "ADDNODES number",
            "int",
            bs.net.addnodes,
            "Add a simulation instance/node",
        ],
        "AIRWAY": [
            "AIRWAY wp/airway",
            "txt",
            bs.traf.airwaycmd,
            "Get info on airway or connections of a waypoint",
        ],
        "ATALT": [
            "acid ATALT alt cmd ",
            "acid,alt,string",
            bs.traf.cond.ataltcmd,
            "When a/c at given altitude , execute a command cmd",
        ],
        "ATDIST": [
            "acid ATDIST pos dist cmd ",
            "acid,latlon,float,string",
            bs.traf.cond.atdistcmd,
            "When a/c passing this distance[nm] to position, execute the command cmd",
        ],
        "ATSPD": [
            "acid ATSPD spd cmd ",
            "acid,spd,string",
            bs.traf.cond.atspdcmd,
            "When a/c reaches given speed, execute a command cmd",
        ],
        "BANK": [
            "BANK acid bankangle[deg]",
            "acid,[float]",
            bs.traf.setbanklim,
            "Set or show bank limit for this vehicle",
        ],
        "BATCH": [
            "BATCH filename",
            "string",
            bs.sim.batch,
            "Start a scenario file as batch simulation",
        ],
        "BLUESKY": ["BLUESKY", "", singbluesky, "Sing"],
        "BENCHMARK": [
            "BENCHMARK [scenfile,time]",
            "[string,time]",
            bs.sim.benchmark,
            "Run benchmark",
        ],
        "BOX": [
            "BOX name,lat,lon,lat,lon,[top,bottom]",
            "txt,latlon,latlon,[alt,alt]",
            lambda name, *coords: areafilter.defineArea(
                name, "BOX", coords[:4], *coords[4:]
            ),
            "Define a box-shaped area",
        ],
        "CALC": [
            "CALC expression",
            "string",
            calculator,
            "Simple in-line math calculator, evaluates expression",
        ],
        "CASMACHTHR": [
            "CASMACHTHR threshold",
            "float",
            aero.casmachthr,
            """Set a threshold below which speeds should be considered as Mach numbers
               in CRE(ATE), ADDWPT, and SPD commands. Set to zero if speeds should
               never be considered as Mach number(e.g., when simulating drones)."""
        ],
        "CD": [
            "CD [path]",
            "[word]",
            setscenpath,
            "Change to a different scenario folder",
        ],
        "CIRCLE": [
            "CIRCLE name,lat,lon,radius,[top,bottom]",
            "txt,latlon,float,[alt,alt]",
            lambda name, *coords: areafilter.defineArea(
                name, "CIRCLE", coords[:3], *coords[3:]
            ),
            "Define a circle-shaped area",
        ],
        "CLRCRECMD": [
            "CLRCRECMD",
            "",
            bs.traf.clrcrecmd,
            "CLRCRECMD will clear CRECMD list of commands a/c creation",
        ],

        "CRE": [
            "CRE acid,type,lat,lon,hdg,alt,spd",
            "txt,txt,latlon,[hdg,alt,spd]",
            bs.traf.cre,
            "Create an aircraft",
        ],
        "CRECMD": [
            "CRECMD cmdline (to be added after a/c id )",
            "string",
            bs.traf.crecmd,
            "Add a command for each aircraft to be issued after creation of aircraft",
        ],
        "CRECONFS": [
            "CRECONFS id, type, targetid, dpsi, cpa, tlos_hor, dH, tlos_ver, spd",
            "txt,txt,acid,hdg,float,time,[alt,time,spd]",
            bs.traf.creconfs,
            "Create an aircraft that is in conflict with 'targetid'",
        ],
        "DATE": [
            "DATE [day,month,year,HH:MM:SS.hh]",
            "[int,int,int,txt]",
            bs.sim.setutc,
            "Set simulation date",
        ],
        "DEFWPT": [
            "DEFWPT wpname,lat,lon,[DELETE/FIX/VOR/DME/NDB/DEL]",
            "txt,latlon,[txt]",
            bs.navdb.defwpt,
            "Define (or delete) a waypoint only for this scenario/run",
        ],
        "DEL": [
            "DEL acid/ALL/WIND/shape",
            "acid/txt,...",
            lambda *a: bs.traf.wind.clear()
            if isinstance(a[0], str) and a[0] == "WIND"
            else areafilter.deleteArea(a[0])
            if isinstance(a[0], str)
            else bs.traf.groups.delgroup(a[0])
            if hasattr(a[0], "groupname")
            else bs.traf.delete(a),
            "Delete command (aircraft, wind, area)",
        ],

        "DIST": [
            "DIST lat0, lon0, lat1, lon1",
            "latlon,latlon",
            distcalc,
            "Distance and direction calculation between two positions",
        ],
        "DOC": [
            "DOC [command]",
            "[txt]",
            bs.scr.show_cmd_doc,
            "Show extended help window for given command, or the main documentation page if no command is given.",
        ],
        "DT": [
            "DT [dt] OR [target,dt]",
            "[float/txt,float]",
            lambda *args: simtime.setdt(*reversed(args)),
            "Set simulation time step",
        ],
        "DTMULT": [
            "DTMULT multiplier",
            "float",
            bs.sim.set_dtmult,
            "Sel multiplication factor for fast-time simulation",
        ],
        "FF": [
            "FF [timeinsec]",
            "[time]",
            bs.sim.fastforward,
            "Fast forward the simulation",
        ],

        "FIXDT": [
            "FIXDT ON/OFF [tend]",
            "onoff,[time]",
            lambda flag, *args: bs.sim.ff(*args) if flag else bs.op(),
            "Legacy function for TMX compatibility",
        ],

        "GROUP": [
            "GROUP [grname, (areaname OR acid,...) ]",
            "[txt,acid/txt,...]",
            bs.traf.groups.group,
            "Add aircraft to a group. OR all aircraft in given area.\n"
            + "Returns list of groups when no argument is passed.\n"
            + "Returns list of aircraft in group when only a groupname is passed.\n"
            + "A group is created when a group with the given name doesn't exist yet.",
        ],
        "HOLD": ["HOLD", "", bs.sim.hold, "Pause(hold) simulation"],
        "IMPLEMENTATION": [
            "IMPLEMENTATION [base, implementation]",
            "[txt,txt]",
            select_implementation,
            "Select an alternate implementation for a Bluesky base class"
        ],
        "LEGEND": [
            "LEGEND label1, ..., labeln",
            "word,...",
            lambda *labels: plotter.legend(labels),
            "Add a legend to the last created plot",
        ],
        "LINE": [
            "LINE name,lat,lon,lat,lon",
            "txt,latlon,latlon",
            lambda name, *coords: areafilter.defineArea(name, "LINE", coords),
            "Draw a line on the radar screen",
        ],
        "LSVAR": [
            "LSVAR path.to.variable",
            "[word]",
            ve.lsvar,
            "Inspect any variable in a bluesky simulation",
        ],
        "MAGVAR": [
            "MAGVAR lat,lon",
            "lat,lon",
            bs.tools.geo.magdeccmd,
            "Show magnetic variation/declination at position",
        ],
        "MCRE": [
            "MCRE n, [type/*, alt/*, spd/*, dest/*]",
            "int,[txt,alt,spd,txt]",
            bs.traf.mcre,
            "Multiple random create of n aircraft in current view",
        ],
        "MOVE": [
            "MOVE acid,lat,lon,[alt,hdg,spd,vspd]",
            "acid,latlon,[alt,hdg,spd,vspd]",
            bs.traf.move,
            "Move an aircraft to a new position",
        ],
        "NOISE": [
            "NOISE [ON/OFF]",
            "[onoff]",
            bs.traf.setnoise,
            "Turbulence/noise switch",
        ],
        "OP": [
            "OP",
            "",
            bs.sim.op,
            "Start/Run simulation or continue after hold"
        ],
        "PLOT": [
            "PLOT [x], y [,dt,colour,figure]",
            "[word,word,float,txt,int]",
            plotter.plot,
            "Create a graph of variables x versus y.",
        ],
        "POLY": [
            "POLY name,[lat,lon,lat,lon, ...]",
            "txt,[latlon,...]",
            lambda name, *coords: areafilter.defineArea(name, "POLY", coords),
            "Define a polygon-shaped area",
        ],
        "POLYALT": [
            "POLYALT name,top,bottom,lat,lon,lat,lon, ...",
            "txt,alt,alt,latlon,...",
            lambda name, top, bottom, *coords: areafilter.defineArea(
                name, "POLYALT", coords, top, bottom
            ),
            "Define a polygon-shaped area in 3D: between two altitudes",
        ],
        "POLYLINE": [
            "POLYLINE name,lat,lon,lat,lon,...",
            "txt,latlon,...",
            lambda name, *coords: areafilter.defineArea(name, "LINE", coords),
            "Draw a multi-segment line on the radar screen",
        ],
        "POS": [
            "POS acid/waypoint",
            "acid/wpt",
            bs.traf.poscommand,
            "Get info on aircraft, airport or waypoint",
        ],
        "REALTIME": [
            "REALTIME [ON/OFF]",
            "[bool]",
            bs.sim.realtime,
            "En-/disable realtime running allowing a variable timestep."],
        "RESET": ["RESET", "", bs.sim.reset, "Reset simulation"],
        "SEED": [
            "SEED value",
            "int",
            bs.sim.setseed,
            "Set seed for all functions using a randomizer (e.g.mcre,noise)",
        ],
        "THR": [
            "THR acid, IDLE/0.0/throttlesetting/1.0/AUTO(default)",
            "acid[,txt]",
            bs.traf.setthrottle,
            "Set throttle or autotothrottle(default)",
        ],
        "TIME": [
            "TIME RUN(default) / HH:MM:SS.hh / REAL / UTC ",
            "[txt]",
            bs.sim.setutc,
            "Set simulated clock time",
        ],
        "TRAIL": [
            "TRAIL ON/OFF, [dt] OR TRAIL acid colour",
            "[acid/bool],[float/txt]",
            bs.traf.trails.setTrails,
            "Toggle aircraft trails on/off",
        ],
        "UNGROUP": [
            "UNGROUP grname, acid",
            "txt,acid,...",
            bs.traf.groups.ungroup,
            "Remove aircraft from a group",
        ],
    }

    #
    # Command synonym dictionary definea equivalent commands globally in stack
    #
    # Actual command definitions: see dictionary in def init(...) below
    #
    synonyms = {
        "ADDAIRWAY": "ADDAWY",
        "AWY": "POS",
        "AIRPORT": "POS",
        "AIRWAYS": "AIRWAY",
        "BANKLIM": "BANK",
        "CHDIR": "CD",
        "COL": "COLOUR",
        "COLOR": "COLOUR",
        "CONTINUE": "OP",
        "CREATE": "CRE",
        "DEBUG": "CALC",
        "DELETE": "DEL",
        "DISP": "SWRAD",
        "FWD": "FF",
        "IMPL": "IMPLEMENTATION",
        "IMPLEMENT": "IMPLEMENTATION",
        "LINES": "POLYLINE",
        "MAGDEC": "MAGVAR",
        "MAGDECL":"MAGVAR",
        "PAUSE": "HOLD",
        "POLYGON": "POLY",
        "POLYLINES": "POLYLINE",
        "PRINT": "ECHO",
        "RT": "REALTIME",
        "RTF": "DTMULT",
        "RUN": "OP",
        "RUNWAYS": "POS",
        "SAVE": "SAVEIC",
        "START": "OP",
        "TRAILS": "TRAIL",
        "VAR": "MAGVAR"
    }

    append_commands(cmddict, synonyms)


def singbluesky():
    webbrowser.open_new("https://youtu.be/aQUlA8Hcv4s")
    return True


def distcalc(lat0, lon0, lat1, lon1):
    try:
        qdr, dist = geo.qdrdist(lat0, lon0, lat1, lon1)
        return True, "QDR = %.2f deg, Dist = %.3f nm" % (qdr % 360.0, dist)
    except:
        return False, "Error in dist calculation."


def setscenpath(newpath):

    if not newpath:
        return False, "Needs an absolute or relative path"

    newpath = Path(newpath)
    # If this is a relative path we need to prefix scenario folder
    if not newpath.is_absolute():
        newpath = bs.resource(settings.scenario_path) / newpath

    if not newpath.is_dir():
        return False, "Error: cannot find path: " + newpath.as_posix()

    # Change path
    settings.scenario_path = newpath

    return True



''' Simulate wind in BlueSky. '''
from numpy import arctan2,degrees,array,sqrt # to allow arrays, their functions and types

from bluesky.tools.aero import kts, ft
from bluesky.core import Entity
from bluesky.stack import command
from .windfield import Windfield


class WindSim(Entity, Windfield, replaceable=True):      
    @command(name='WIND')
    def add(self, lat: 'lat', lon: 'lon', *winddata: 'float/alt'):
        """ Define a wind vector as part of the 2D or 3D wind field.
        
            Arguments:
            - lat/lon: Horizonal position to define wind vector(s)
            - winddata: 
              - If the wind at this location is independent of altitude
                winddata has two elements:
                - direction [degrees]
                - speed (magnitude) [knots]
              - If the wind varies with altitude winddata has three elements:
                - altitude [ft]
                - direction [degrees]
                - speed (magnitude) [knots]
                In this case, repeating combinations of alt/dir/spd can be provided
                to specify wind at multiple altitudes.
        """
        ndata = len(winddata)

        # No altitude or just one: same wind for all altitudes at this position
        if ndata == 2 or (ndata == 3 and winddata[0] is None): # only one point, ignore altitude
            if winddata[-2] is None or winddata[-1] is None:
               return False, "Wind direction and speed needed."

            self.addpoint(lat,lon,winddata[-2],winddata[-1]*kts)

        # More than one altitude is given
        elif ndata >= 3:
            windarr = array(winddata)
            dirarr = windarr[1::3]
            spdarr = windarr[2::3] * kts
            altarr = windarr[0::3] * ft

            self.addpoint(lat,lon,dirarr,spdarr,altarr)
            
        elif winddata.count("DEL") > 0:
            self.clear()

        else:# Something is wrong
            return False, "Winddata not recognized"

        return True

    @command(name='GETWIND')
    def get(self, lat: 'lat', lon: 'lon', alt: 'alt'=None):
        """ Get wind at a specified position (and optionally at altitude) 
        
            Arguments:
            - lat, lon: Horizontal position where wind should be determined [deg]
            - alt: Altitude at which wind should be determined [ft]
        """
        vn,ve = self.getdata(lat,lon,alt)

        wdir = (degrees(arctan2(ve,vn)) + 180.) % 360.
        wspd = sqrt(vn * vn + ve * ve)

        txt  = "WIND AT %.5f, %.5f: %03d/%d" % (lat,lon,round(wdir),round(wspd/kts))

        return True, txt


""" Geovectoring: for an area define an allowed interval for each component
    of the 3D speed vector (gs,trk,vs):

    first argument is the are on whuch the geovector shoudl apply

    Geovector is defined as:
    (  [ gsmin,gsmax ]  ) [kts]
    (  [trkmin,trkmax]  ) [deg]
    (  [ vsmin,vsmax ]  ) [fpm]
"""
import numpy as np

from bluesky import traf  #, settings, navdb, traf, sim, scr, tools
from bluesky.tools import areafilter
from bluesky.tools.aero import vtas2cas,ft
from bluesky.tools.misc import degto180


# A dictionary of areas with a geovector specification
geovecs = dict()


class GeoVector:
    ''' Geovector specification class. '''
    def __init__(self, areaname, gsmin, gsmax,
                 trkmin, trkmax,
                 vsmin, vsmax):
        self.areaname = areaname
        self.gsmin = gsmin
        self.gsmax = gsmax
        self.trkmin = trkmin
        self.trkmax = trkmax
        self.vsmin = vsmin
        self.vsmax = vsmax
        self.previnside = set()
        self.prevstatus = dict()

    def __str__(self):
        ''' Pretty printed geovector information. '''
        return f'gs: {self.gsmin}-{self.gsmax} [m/s], ' + \
            f'{self.trkmin}-{self.trkmax} trk[deg], ' + \
                f'{self.vsmin}-{self.vsmax} vs[m/s]'


def init_plugin():

    # Create an empty geovector list
    reset()

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name'      : 'GEOVECTOR',
        'plugin_type'      : 'sim',
        'update_interval'  :  1.0,

        # The update function is called after traffic is updated. Use this if you
        # want to do things as a result of what happens in traffic. If you need to
        # something before traffic is updated please use preupdate.
        'update':          update,

        # The preupdate function is called before traffic is updated. Use this
        # function to provide settings that need to be used by traffic in the current
        # timestep. Examples are ASAS, which can give autopilot commands to resolve
        # a conflict.
        'preupdate':       preupdate,

        # Reset all geovectors
        'reset':         reset
        }

    # Add two commands: GEOVECTOR to define a geovector for an area
    stackfunctions = {
        # Defining a geovector
        'GEOVECTOR': [
            'GEOVECTOR area,[gsmin,gsmax,trkmin,trkmax,vsmin,vsmax]',
            'txt,[spd,spd,hdg,hdg,vspd,vspd]',
            defgeovec,
            'Define a geovector for an area defined with the BOX,POLY(ALT) area commands']
        ,
        # Delete a geovector (same effect as using a geovector without  any values
        'DELGEOVECTOR': [
            'DELGEOVECTOR area',
            'txt',
            delgeovec,
            'Remove geovector from the area ']
        }
    # init_plugin() should always return these two dicts.
    return config, stackfunctions


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

def preupdate(): # To be safe preupdate is used iso update
    applygeovec()
    return

def applygeovec():
    # Apply each geovector
    for areaname, vec in geovecs.items():
        if areafilter.hasArea(areaname):
            swinside  = areafilter.checkInside(areaname, traf.lat, traf.lon, traf.alt)

            insids = set(np.array(traf.id)[swinside])
            newids = insids - vec.previnside
            delids = vec.previnside - insids
            # Store LNAV/VNAV status of new aircraft
            for acid in newids:
                idx = traf.id2idx(acid)
                vec.prevstatus[acid] = [traf.swlnav[idx], traf.swvnav[idx]]

            # Revert aircraft who have exited the geovectored area to their original status
            for acid in delids:
                idx = traf.id2idx(acid)
                if idx >= 0:
                    traf.swlnav[idx], traf.swvnav[idx] = vec.prevstatus.pop(acid)
            
            vec.previnside = insids
            # -----Ground speed limiting
            # For now assume no wind:  so use tas as gs
            if vec.gsmin is not None:
                casmin = vtas2cas(np.ones(traf.ntraf) * vec.gsmin, traf.alt)
                usemin = traf.selspd < casmin
                traf.selspd[swinside & usemin] = casmin[swinside & usemin]
                traf.swvnav[swinside & usemin] = False

            if vec.gsmax is not None:
                casmax = vtas2cas(np.ones(traf.ntraf) * vec.gsmax, traf.alt)
                usemax = traf.selspd > casmax
                traf.selspd[swinside & usemax] = casmax[swinside & usemax]
                traf.swvnav[swinside & usemax] = False

            #------ Limit Track(so hdg)
            # Max track interval is 180 degrees to avoid ambiguity of what is inside the interval

            if None not in [vec.trkmin, vec.trkmax]:
                # Use degto180 to avodi problems for e.g interval [350,30]
                usemin = swinside & (degto180(traf.trk - vec.trkmin) < 0) # Left of minimum
                usemax = swinside & (degto180(traf.trk - vec.trkmax) > 0) # Right of maximum

                #print(usemin,usemax)
                traf.swlnav[swinside & (usemin | usemax)] = False

                traf.ap.trk[swinside & usemin] = vec.trkmin
                traf.ap.trk[swinside & usemax] = vec.trkmax

            # -----Ground speed limiting
            # For now assume no wind:  so use tas as gs
            if vec.vsmin is not None:
                traf.selvs[swinside & (traf.vs < vec.vsmin)] = vec.vsmin
                traf.swvnav[swinside & (traf.vs < vec.vsmin)] = False
                # Activate V/S mode by using a slightly higher altitude than current values
                traf.selalt[swinside & (traf.vs < vec.vsmin)] = traf.alt[swinside & (traf.vs < vec.vsmin)] + \
                                                            np.sign(vec.vsmin)*200.*ft

            if vec.vsmax is not None:
                traf.selvs[swinside & (traf.vs > vec.vsmax)] = vec.vsmax
                traf.swvnav[swinside & (traf.vs < vec.vsmax)] = False
                # Activate V/S mode by using a slightly higher altitude than current values
                traf.selalt[swinside & (traf.vs > vec.vsmax)] = traf.alt[swinside & (traf.vs > vec.vsmax)] + \
                                                            np.sign(vec.vsmax)*200.*ft

    return

def update(): # Not used
    return

def reset():
    geovecs.clear() # [[area,gsmin,gsmax,trkin,trkmax,vsmin,vsmax]]
    return

### Other functions of your plug-in

def defgeovec(areaname="", spdmin=None, spdmax=None, trkmin=None, trkmax=None, vspdmin=None, vspdmax=None):
    #print ("defgeovec input=",areaname,gsmin,gsmax,trkmin,trkmax,vspdmin,vspdmax,sep="|")

    # We need an area to do anything
    if not areaname:
        return False, "We need an area"

    if not (spdmin or spdmax or (trkmin and trkmax) or vspdmin or vspdmax):
        if areaname in geovecs:
            return True, f'{areaname} uses {geovecs[areaname]}'
        return False, f'No geovector found for {areaname}'

    # Remove old geovector for this area (if it exists)
    geovecs.pop(areaname, None)

    # Only add it if an interval is given
    if spdmin or spdmax or (trkmin and trkmax) or vspdmin or vspdmax:
        # Allow wrong order of min and max (note for hdg, as 355,10 is valid range!)
        if None not in [spdmin, spdmax]:
            gsmin = min(spdmin,spdmax)
            gsmax = max(spdmin,spdmax)
        else:
            gsmin = spdmin
            gsmax = spdmax

        if None not in [vspdmin, vspdmax]:
            vsmin = min(vspdmin,vspdmax)
            vsmax = max(vspdmin,vspdmax)
        else:
            vsmin = vspdmin
            vsmax = vspdmax

        # Add geovector to the dict of geovectors
        geovecs[areaname] = GeoVector(areaname,
                        gsmin, gsmax,
                        trkmin,trkmax,
                        vsmin, vsmax)

    return True

def delgeovec(area=""):
    # Delete geovector specification for area if it exists
    found = geovecs.pop(area)
    if not found:
        return False, f'No geovector found for {area}'

    return True, f'Deleted geovector specification for {area}'



''' BlueSky UI classes and functions. '''
# from bluesky.ui.qtgl.glhelpers import (gl, RenderObject, RenderTarget, RenderWidget,
#     ShaderProgram, ShaderSet, VertexArrayObject, GLBuffer, UniformBufferObject, Font,
#     Text, Texture, Circle, Rectangle)


""" Autopilot Implementation."""
from math import sqrt, atan
import numpy as np
from collections.abc import Collection
import bluesky as bs
from bluesky import stack
from bluesky.tools import geo
from bluesky.tools.misc import degto180
from bluesky.tools.position import txt2pos
from bluesky.tools.aero import ft, nm, fpm, vcasormach2tas, vcas2tas, tas2cas, cas2tas, g0
from bluesky.core import Entity
from .route import Route

# debug
from inspect import stack as callstack
from bluesky.tools.datalog import crelog

bs.settings.set_variable_defaults(fms_dt=10.5)


class Autopilot(Entity, replaceable=True):
    ''' BlueSky Autopilot implementation. '''
    def __init__(self):
        super().__init__()

        # Standard self.steepness for descent
        self.steepness = 3000. * ft / (10. * nm)

        # From here, define object arrays
        with self.settrafarrays():

            # FMS directions
            self.trk = np.array([])
            self.spd = np.array([])
            self.tas = np.array([])
            self.alt = np.array([])
            self.vs  = np.array([])

            # VNAV variables
            self.swtoc    = np.array([])  # ToC switch to switch on VNAV Top of Climb logic (default value True)
            self.swtod    = np.array([])  # ToD switch to switch on VNAV Top of Descent logic (default value True)

            self.dist2vs  = np.array([])  # distance from coming waypoint to TOD
            self.dist2accel = np.array([]) # Distance to go to acceleration(decelaration) for turn next waypoint [nm]

            self.swvnavvs = np.array([])  # whether to use given VS or not
            self.vnavvs   = np.array([])  # vertical speed in VNAV


            # LNAV variables
            self.qdr2wp      = np.array([]) # Direction to waypoint from the last time passing was checked
                                            # to avoid 180 turns due to updated qdr shortly before passing wp
            self.dist2wp     = np.array([]) # [m] Distance to active waypoint
            self.qdrturn     = np.array([]) # qdr to next turn]
            self.dist2turn   = np.array([]) # Distance to next turn [m]

            self.inturn = np.array([]) # If we're in a turn maneuver or not

            # Traffic navigation information
            self.orig = []  # Four letter code of origin airport
            self.dest = []  # Four letter code of destination airport

            # Default values
            self.bankdef = np.array([])  # nominal bank angle, [radians]
            self.vsdef = np.array([]) # [m/s]default vertical speed of autopilot
            self.cruisespd = np.array([]) # Cruise speed to apply when exiting a turn
            
            # Currently used roll/bank angle [rad]
            self.turnphi = np.array([])  # [rad] bank angle setting of autopilot

            # Route objects
            self.route = []


        self.idxreached = []    # List indices of aircraft who have reached their active waypoint

    def create(self, n=1):
        super().create(n)

        # FMS directions
        self.trk[-n:] = bs.traf.trk[-n:]
        self.tas[-n:] = bs.traf.tas[-n:]
        self.alt[-n:] = bs.traf.alt[-n:]
        self.vs[-n:]  = -999

        # Default ToC/ToD logic on
        self.swtoc[-n:] = True
        self.swtod[-n:] = True

        # VNAV Variables
        self.dist2vs[-n:] = -999.
        self.dist2accel[-n:] = -999.  # Distance to go to acceleration(decelaration) for turn next waypoint [nm]


        # LNAV variables
        self.qdr2wp[-n:] = -999.   # Direction to waypoint from the last time passing was checked
        self.dist2wp[-n:]  = -999. # Distance to go to next waypoint [nm]

        # Traffic performance data
        #(temporarily default values)
        self.vsdef[-n:] = 1500. * fpm   # default vertical speed of autopilot
        self.bankdef[-n:] = np.radians(25.) # default bank angle
        self.cruisespd[-n:] = -999. # default cruise speed

        # Route objects
        for ridx, acid in enumerate(bs.traf.id[-n:]):
            self.route[ridx - n] = Route(acid)

    def wppassingcheck(self, qdr, dist): # qdr [deg], dist [m[
        """
        The actwp is the interface between the list of waypoint data in the route object and the autopilot guidance
        when LNAV is on (heading) and optionally VNAV is on (spd & altitude)

        actwp data contains traffic arrays, to allow vectorizing the guidance logic.

        Waypoint switching (just like the adding, deletion in route) are event driven commands and
        therefore not vectorized as they occur rarely compared to the guidance.

        wppassingcheck contains the waypoint switching function:
        - Check which aircraft i have reached their active waypoint
        - Reached function return list of indices where reached logic is True
        - Get the waypoint data to the actwp (active waypoint data)
        - Shift waypoint (last,next etc.) data for aircraft i where necessary
        - Shift and maintain data (see last- and next- prefix in varubale name) e.g. to continue a special turn
        - Prepare some VNAV triggers along the new leg for the VNAV profile (where to start descent/climb)
        """

        # Get list of indices of aircraft which have reached their active waypoint
        # This vectorized function checks the passing of the waypoint using a.o. the current turn radius
        self.idxreached = bs.traf.actwp.reached(qdr, dist)

        # For the one who have reached their active waypoint, update vectorized leg data for guidance
        for i in self.idxreached:

            #debug commands to check VNAV state while passing waypoint
            #print("Passing waypoint",bs.traf.ap.route[i].wpname[bs.traf.ap.route[i].iactwp])
            #print("dist2wp,dist2vs",self.dist2wp[i]/nm,self.dist2vs[i]/nm) # distance to wp & distance to ToD/ToC

            # Save current wp speed for use on next leg when we pass this waypoint
            # VNAV speeds are always FROM-speeds, so we accelerate/decellerate at the waypoint
            # where this speed is specified, so we need to save it for use now
            # before getting the new data for the next waypoint

            # Get speed for next leg from the waypoint we pass now and set as active spd
            bs.traf.actwp.spd[i]    = bs.traf.actwp.nextspd[i]
            bs.traf.actwp.spdcon[i] = bs.traf.actwp.nextspd[i]

            # Execute stack commands for the still active waypoint, which we pass now
            self.route[i].runactwpstack()
            
            # Get next wp, if there still is one
            if not bs.traf.actwp.swlastwp[i]:
                lat, lon, alt, bs.traf.actwp.nextspd[i], \
                bs.traf.actwp.xtoalt[i], toalt, \
                    bs.traf.actwp.xtorta[i], bs.traf.actwp.torta[i], \
                    lnavon, flyby, flyturn, turnrad, turnspd, turnhdgr, turnbank,\
                    bs.traf.actwp.next_qdr[i], bs.traf.actwp.swlastwp[i] =      \
                    self.route[i].getnextwp()  # [m] note: xtoalt,nextaltco are in meters


                bs.traf.actwp.nextturnlat[i], bs.traf.actwp.nextturnlon[i], \
                bs.traf.actwp.nextturnspd[i], bs.traf.actwp.nextturnrad[i], \
                bs.traf.actwp.nextturnhdgr[i],bs.traf.actwp.nextturnidx[i] = \
                    self.route[i].getnextturnwp()

            else:
                # Prevent trying to activate the next waypoint when it was already the last waypoint
                # In case of end of route/no more waypoints: switch off LNAV using the lnavon
                bs.traf.swlnav[i] = False
                bs.traf.swvnav[i] = False
                bs.traf.swvnavspd[i] = False
                continue # Go to next a/c which reached its active waypoint

            # Special turns: specified by one or two of the following variables:
            # Turn speed, turn radius, turn bank angle, turn rate
            # if flyturn:
            #     # First situation, if turn rate is specified
            #     if turnspd<=0.:
            #         turnspd = bs.traf.tas[i]

            #     # Heading rate overrides turnrad
            #     if turnhdgr>0:
            #         turnrad = bs.traf.tas[i]*360./(2*np.pi*turnhdgr)

            #     # Use last turn radius for bank angle in current turn
            #     if bs.traf.actwp.turnrad[i] > 0.:
            #         self.turnphi[i] = atan(bs.traf.actwp.turnspd[i]*bs.traf.actwp.turnspd[i]/ \
            #                                (bs.traf.actwp.turnrad[i]*g0)) # [rad]
            #     else:
            #         self.turnphi[i] = 0.0  # [rad] or leave untouched???

            # else:
            #     self.turnphi[i] = 0.0  #[rad] or leave untouched???



            # Check LNAV switch returned by getnextwp
            # Switch off LNAV if it failed to get next wpdata
            if not lnavon and bs.traf.swlnav[i]:
                bs.traf.swlnav[i] = False
                # Last wp: copy last wp values for alt and speed in autopilot
                if bs.traf.swvnavspd[i] and bs.traf.actwp.nextspd[i]>= 0.0:
                    bs.traf.selspd[i] = bs.traf.actwp.nextspd[i]

            # In case of no LNAV, do not allow VNAV mode to be active
            bs.traf.swvnav[i] = bs.traf.swvnav[i] and bs.traf.swlnav[i]

            bs.traf.actwp.lat[i] = lat  # [deg]
            bs.traf.actwp.lon[i] = lon  # [deg]
            # 1.0 in case of fly by, else fly over
            bs.traf.actwp.flyby[i] = int(flyby)

            # Update qdr and turndist for this new waypoint for ComputeVNAV
            qdr[i], distnmi = geo.qdrdist(bs.traf.lat[i], bs.traf.lon[i],
                                          bs.traf.actwp.lat[i], bs.traf.actwp.lon[i])

            #dist[i] = distnmi * nm
            self.dist2wp[i] = distnmi*nm

            bs.traf.actwp.curlegdir[i] = qdr[i]
            bs.traf.actwp.curleglen[i] = self.dist2wp[i]

            # User has entered an altitude for the new waypoint
            if alt >= -0.01: # positive alt on this waypoint means altitude constraint
                bs.traf.actwp.nextaltco[i] = alt  # [m]
                bs.traf.actwp.xtoalt[i] = 0.0
            else:
                bs.traf.actwp.nextaltco[i] = toalt  # [m]

            #if not bs.traf.swlnav[i]:
            #    bs.traf.actwp.spd[i] = -997.

            # VNAV spd mode: use speed of this waypoint as commanded speed
            # while passing waypoint and save next speed for passing next wp
            # Speed is now from speed! Next speed is ready in wpdata
            if bs.traf.swvnavspd[i] and bs.traf.actwp.spd[i]>= 0.0:
                    bs.traf.selspd[i] = bs.traf.actwp.spd[i]

            # Update turndist so ComputeVNAV works, is there a next leg direction or not?
            if bs.traf.actwp.next_qdr[i] < -900.:
                local_next_qdr = qdr[i]
            else:
                local_next_qdr = bs.traf.actwp.next_qdr[i]

            # Calculate turn dist (and radius which we do not use now, but later) now for scalar variable [i]
            bs.traf.actwp.turndist[i], turnrad, turnspd, turnbank, turnhdgr = \
                bs.traf.actwp.calcturn(i, bs.traf.tas[i], qdr[i], 
                                       local_next_qdr, turnbank, 
                                       turnrad,turnspd,turnhdgr, flyturn, flyby)  # update turn distance for VNAV

            # Get flyturn switches and data
            bs.traf.actwp.flyturn[i]      = flyturn
            bs.traf.actwp.turnrad[i]      = turnrad
            bs.traf.actwp.turnspd[i]      = turnspd
            bs.traf.actwp.turnhdgr[i]     = turnhdgr
            self.turnphi[i] = np.deg2rad(turnbank)

            # Pass on whether currently flyturn mode:
            # at beginning of leg,c copy tonextwp to lastwp
            # set next turn False
            bs.traf.actwp.turnfromlastwp[i] = bs.traf.actwp.turntonextwp[i]
            bs.traf.actwp.turntonextwp[i]   = False

            # Keep both turning speeds: turn to leg and turn from leg
            bs.traf.actwp.oldturnspd[i]  = bs.traf.actwp.turnspd[i] # old turnspd, turning by this waypoint
            if bs.traf.actwp.flyturn[i]:
                bs.traf.actwp.turnspd[i] = turnspd                  # new turnspd, turning by next waypoint
            else:
                bs.traf.actwp.turnspd[i] = -990.

            # Reduce turn dist for reduced turnspd
            if bs.traf.actwp.flyturn[i] and bs.traf.actwp.turnrad[i]<0.0 and bs.traf.actwp.turnspd[i]>=0.:
                turntas = cas2tas(bs.traf.actwp.turnspd[i], bs.traf.alt[i])
                bs.traf.actwp.turndist[i] = bs.traf.actwp.turndist[i]*turntas*turntas/(bs.traf.tas[i]*bs.traf.tas[i])

            # VNAV = FMS ALT/SPD mode incl. RTA
            self.ComputeVNAV(i, toalt, bs.traf.actwp.xtoalt[i], bs.traf.actwp.torta[i],
                             bs.traf.actwp.xtorta[i])

        

        # End of reached-loop: the per waypoint i switching loop

        # Update qdr2wp with up-to-date qdr, now that we have checked passing wp
        self.qdr2wp = qdr%360.

        # Continuous guidance when speed constraint on active leg is in update-method

        # If still an RTA in the route and currently no speed constraint
        for iac in np.where((bs.traf.actwp.torta > -99.)*(bs.traf.actwp.spdcon<0.0))[0]:
            iwp = bs.traf.ap.route[iac].iactwp
            if bs.traf.ap.route[iac].wprta[iwp]>-99.:

                 # For all a/c flying to an RTA waypoint, recalculate speed more often
                dist2go4rta = geo.kwikdist(bs.traf.lat[iac],bs.traf.lon[iac],
                                           bs.traf.actwp.lat[iac],bs.traf.actwp.lon[iac])*nm \
                               + bs.traf.ap.route[iac].wpxtorta[iwp] # last term zero for active wp rta

                # Set bs.traf.actwp.spd to rta speed, if necessary
                self.setspeedforRTA(iac,bs.traf.actwp.torta[iac],dist2go4rta)

                # If VNAV speed is on (by default coupled to VNAV), use it for speed guidance
                if bs.traf.swvnavspd[iac] and bs.traf.actwp.spd[iac]>=0.0:
                     bs.traf.selspd[iac] = bs.traf.actwp.spd[iac]

    def update(self):
        # FMS LNAV mode:
        # qdr[deg],distinnm[nm]
        qdr, distinnm = geo.qdrdist(bs.traf.lat, bs.traf.lon,
                                    bs.traf.actwp.lat, bs.traf.actwp.lon)  # [deg][nm])

        self.qdr2wp  = qdr
        self.dist2wp = distinnm*nm  # Conversion to meters

        # Check possible waypoint shift. Note: qdr, dist2wp will be updated accordingly in case of wp switch
        self.wppassingcheck(qdr, self.dist2wp) # Updates self.qdr2wp when necessary

        #================= Continuous FMS guidance ========================

        # Note that the code below is vectorized, with traffic arrays, so for all aircraft
        # ComputeVNAV and inside waypoint loop of wppassingcheck, it was scalar (per a/c with index i)

        # VNAV altitude guidance logic (using the variables prepared by ComputeVNAV when activating waypoint)

        # First question is:
        # - Can we please we start to descend or to climb?
        #
        # The variable dist2vs indicates the distance to the active waypoint where we should start our climb/descend
        # Only use this logic if there is a valid next altitude constraint (nextaltco).
        #
        # Well, when Top of Descent (ToD) switch is on, descend as late as possible,
        # But when Top of Climb switch is on or off, climb as soon as possible, only difference is steepness used in ComputeVNAV
        # to calculate bs.traf.actwp.vs

        startdescorclimb = (bs.traf.actwp.nextaltco>=-0.1) * \
                           np.logical_or((bs.traf.alt>bs.traf.actwp.nextaltco) *\
                                         np.logical_or((self.dist2wp < self.dist2vs+bs.traf.actwp.turndist),
                                                       (np.logical_not(self.swtod))),
                                         bs.traf.alt<bs.traf.actwp.nextaltco)

        # print("self.dist2vs =",self.dist2vs)

        # If not lnav:Climb/descend if doing so before lnav/vnav was switched off
        #    (because there are no more waypoints). This is needed
        #    to continue descending when you get into a conflict
        #    while descending to the destination (the last waypoint)
        #    Use 0.1 nm (185.2 m) circle in case turndist might be zero
        self.swvnavvs = bs.traf.swvnav * np.where(bs.traf.swlnav, startdescorclimb,
                                        self.dist2wp <= np.maximum(0.1*nm,bs.traf.actwp.turndist))

        # Recalculate V/S based on current altitude and distance to next alt constraint
        # How much time do we have before we need to descend?
        # Now done in ComputeVNAV
        # See ComputeVNAV for bs.traf.actwp.vs calculation

        self.vnavvs  = np.where(self.swvnavvs, bs.traf.actwp.vs, self.vnavvs)
        #was: self.vnavvs  = np.where(self.swvnavvs, self.steepness * bs.traf.gs, self.vnavvs)

        # self.vs = np.where(self.swvnavvs, self.vnavvs, self.vsdef * bs.traf.limvs_flag)
        # for VNAV use fixed V/S and change start of descent
        selvs = np.where(abs(bs.traf.selvs) > 0.1, bs.traf.selvs, self.vsdef) # m/s
        self.vs  = np.where(self.swvnavvs, self.vnavvs, selvs)
        self.alt = np.where(self.swvnavvs, bs.traf.actwp.nextaltco, bs.traf.selalt)

        # When descending or climbing in VNAV also update altitude command of select/hold mode
        bs.traf.selalt = np.where(self.swvnavvs,bs.traf.actwp.nextaltco,bs.traf.selalt)

        # LNAV commanded track angle
        self.trk = np.where(bs.traf.swlnav, self.qdr2wp, self.trk)

        # FMS speed guidance: anticipate accel/decel distance for next leg or turn

        # Calculate actual distance it takes to decelerate/accelerate based on two cases: turning speed (decel)

        # Normally next leg speed (actwp.spd) but in case we fly turns with a specified turn speed
        # use the turn speed

        # Is turn speed specified and are we not already slow enough? We only decelerate for turns, not accel.
        turntas       = np.where(bs.traf.actwp.nextturnspd>0.0, vcas2tas(bs.traf.actwp.nextturnspd, bs.traf.alt),
                                 -1.0+0.*bs.traf.tas)
        
        # Switch is now whether the aircraft has any turn waypoints
        swturnspd     = bs.traf.actwp.nextturnidx >= 0
        turntasdiff   = np.maximum(0.,(bs.traf.tas - turntas)*(turntas>0.0))

        # t = (v1-v0)/a ; x = v0*t+1/2*a*t*t => dx = (v1*v1-v0*v0)/ (2a)
        dxturnspdchg = distaccel(turntas,bs.traf.tas, bs.traf.perf.axmax)

        # Decelerate or accelerate for next required speed because of speed constraint or RTA speed
        # Note that because nextspd comes from the stack, and can be either a mach number or
        # a calibrated airspeed, it can only be converted from Mach / CAS [kts] to TAS [m/s]
        # once the altitude is known.
        nexttas = vcasormach2tas(bs.traf.actwp.nextspd, bs.traf.alt)
#
        dxspdconchg = distaccel(bs.traf.tas, nexttas, bs.traf.perf.axmax)

        qdrturn, dist2turn = geo.qdrdist(bs.traf.lat, bs.traf.lon,
                                        bs.traf.actwp.nextturnlat, bs.traf.actwp.nextturnlon)

        self.qdrturn = qdrturn
        dist2turn = dist2turn * nm

        # Where we don't have a turn waypoint, as in turn idx is negative, then put distance
        # as Earth circumference.
        self.dist2turn = np.where(bs.traf.actwp.nextturnidx >= 0, dist2turn, 40075000)

        # Check also whether VNAVSPD is on, if not, SPD SEL has override for next leg
        # and same for turn logic
        usenextspdcon = (self.dist2wp < dxspdconchg)*(bs.traf.actwp.nextspd>-990.) * \
                            bs.traf.swvnavspd*bs.traf.swvnav*bs.traf.swlnav

        useturnspd = np.logical_or(bs.traf.actwp.turntonextwp,
                                   (self.dist2turn < (dxturnspdchg*1.1+np.maximum(bs.traf.actwp.turndist,bs.traf.actwp.nextturnrad)))) * \
                                        swturnspd*bs.traf.swvnavspd*bs.traf.swvnav*bs.traf.swlnav

        # Hold turn mode can only be switched on here, cannot be switched off here (happeps upon passing wp)
        bs.traf.actwp.turntonextwp = bs.traf.swlnav*np.logical_or(bs.traf.actwp.turntonextwp,useturnspd)

        # Which CAS/Mach do we have to keep? VNAV, last turn or next turn?
        oncurrentleg = (abs(degto180(bs.traf.trk - qdr)) < 2.0) # [deg]
        inoldturn    = (bs.traf.actwp.oldturnspd > 0.) * np.logical_not(oncurrentleg)

        # Avoid using old turning speeds when turning of this leg to the next leg
        # by disabling (old) turningspd when on leg
        bs.traf.actwp.oldturnspd = np.where(oncurrentleg*(bs.traf.actwp.oldturnspd>0.), -998.,
                                            bs.traf.actwp.oldturnspd)

        # turnfromlastwp can only be switched off here, not on (latter happens upon passing wp)
        bs.traf.actwp.turnfromlastwp = np.logical_and(bs.traf.actwp.turnfromlastwp,inoldturn)

        # Select speed: turn sped, next speed constraint, or current speed constraint
        bs.traf.selspd = np.where(useturnspd,bs.traf.actwp.nextturnspd,
                                  np.where(usenextspdcon, bs.traf.actwp.nextspd,
                                           np.where((bs.traf.actwp.spdcon>=0)*bs.traf.swvnavspd,bs.traf.actwp.spd,
                                                                            bs.traf.selspd)))

        # Temporary override speed when still in old turn
        bs.traf.selspd = np.where(np.logical_and(inoldturn*(bs.traf.actwp.oldturnspd>0.)*bs.traf.swvnavspd*bs.traf.swvnav*bs.traf.swlnav,
                                    np.logical_not(useturnspd)),
                                  bs.traf.actwp.oldturnspd,bs.traf.selspd)

        # Before updating inturn, save the old inturn
        oldinturn = self.inturn
        
        self.inturn = np.logical_or(useturnspd,inoldturn)
        
        # Turn was exited this time step
        justexitedturn = np.logical_and(oldinturn, np.logical_not(self.inturn))
        
        # Apply the cruise speed if the past or next waypoint doesn't have a speed constraint 
        # and if there is actually a cruise speed to apply
        usecruisespd = np.logical_and.reduce((self.cruisespd > 0,
                                              bs.traf.actwp.spd < 0,
                                              np.logical_not(usenextspdcon),
                                              justexitedturn))
        
        bs.traf.selspd = np.where(usecruisespd, self.cruisespd, bs.traf.selspd)

        # Below crossover altitude: CAS=const, above crossover altitude: Mach = const
        self.tas = vcasormach2tas(bs.traf.selspd, bs.traf.alt)

    def ComputeVNAV(self, idx, toalt, xtoalt, torta, xtorta):
        """
        This function to do VNAV (and RTA) calculations is only called only once per leg for one aircraft idx.
        If:
         - switching to next waypoint
         - when VNAV is activated
         - when a DIRECT is given

        It prepares the profile of this leg using the the current altitude and the next altitude constraint (nextaltco).
        The distance to the next altitude constraint is given by xtoalt [m] after active waypoint.

        Options are (classic VNAV logic, swtoc and swtod True):
        - no altitude constraint in the future, do nothing
        - Top of CLimb logic (swtoc=True): if next altitude constrain is baove us, climb as soon as possible with default steepness
        - Top of Descent Logic (swtod =True) Use ToD logic: descend as late aspossible, based on
          steepness. Prepare a ToD somewhere on the leg if necessary based on distance to next altitude constraint.
          This is done by calculating distance to next waypoint where descent should start

        Alternative logic (e.g. for UAVs or GA):
        - swtoc=False and next alt co is above us, climb with the angle/steepness needed to arrive at the altitude at
        the waypoint with the altitude constraint (xtoalt m after active waypoint)
        - swtod=False and next altco is below us, descend with the angle/steepness needed to arrive at at the altitude at
        the waypoint with the altitude constraint (xtoalt m after active waypoint)

        Output if this function:
        self.dist2vs = distance 2 next waypoint where climb/descent needs to activated
        bs.traf.actwp.vs =  V/S to be used during climb/descent part, so when dist2wp<dist2vs [m] (to next waypoint)
        """

        #print ("ComputeVNAV for",bs.traf.id[idx],":",toalt/ft,"ft  ",xtoalt/nm,"nm")
        #print("Called by",callstack()[1].function)

        # Check  whether active waypoint speed needs to be adjusted for RTA
        # sets bs.traf.actwp.spd, if necessary
        # debug print("xtorta+legdist =",(xtorta+legdist)/nm)
        self.setspeedforRTA(idx, torta, xtorta + self.dist2wp[idx])  # all scalar

        # Check if there is a target altitude and VNAV is on, else return doing nothing
        if toalt < 0 or not bs.traf.swvnav[idx]:
            self.dist2vs[idx] = -999999. #dist to next wp will never be less than this, so VNAV will do nothing
            return

        # So: somewhere there is an altitude constraint ahead
        # Compute proper values for bs.traf.actwp.nextaltco, self.dist2vs, self.alt, bs.traf.actwp.vs
        # Descent VNAV mode (T/D logic)
        #
        # xtoalt  =  distance to go to next altitude constraint at a waypoint in the route
        #            (could be beyond next waypoint) [m]
        #
        # toalt   = altitude at next waypoint with an altitude constraint
        #
        # dist2vs = autopilot starts climb or descent when the remaining distance to next waypoint
        #           is this distance
        #
        #
        # VNAV Guidance principle:
        #
        #
        #                          T/C------X---T/D
        #                           /    .        \
        #                          /     .         \
        #       T/C----X----.-----X      .         .\
        #       /           .            .         . \
        #      /            .            .         .  X---T/D
        #     /.            .            .         .        \
        #    / .            .            .         .         \
        #   /  .            .            .         .         .\
        # pos  x            x            x         x         x X
        #
        #
        #  X = waypoint with alt constraint  x = Wp without prescribed altitude
        #
        # - Ignore and look beyond waypoints without an altitude constraint
        # - Climb as soon as possible after previous altitude constraint
        #   and climb as fast as possible, so arriving at alt earlier is ok
        # - Descend at the latest when necessary for next altitude constraint
        #   which can be many waypoints beyond current actual waypoint
        epsalt = 2.*ft # deadzone
        #
        if bs.traf.alt[idx] > toalt + epsalt:
            # Stop potential current climb (e.g. due to not making it to previous altco)
            # then stop immediately, as in: do not make it worse.
            if bs.traf.vs[idx]>0.0001:
                self.vnavvs[idx] = 0.0
                self.alt[idx] = bs.traf.alt[idx]
                if bs.traf.swvnav[idx]:
                    bs.traf.selalt[idx] = bs.traf.alt[idx]

            # Descent modes: VNAV (= swtod/Top of Descent logic) or aiming at next alt constraint

            # Calculate max allowed altitude at next wp (above toalt)
            bs.traf.actwp.nextaltco[idx] = toalt  # [m] next alt constraint
            bs.traf.actwp.xtoalt[idx]    = xtoalt # [m] distance to next alt constraint measured from next waypoint


            # VNAV ToD logic
            if self.swtod[idx]:
                # Get distance to waypoint
                self.dist2wp[idx] = nm*geo.kwikdist(bs.traf.lat[idx], bs.traf.lon[idx],
                                                 bs.traf.actwp.lat[idx],
                                                 bs.traf.actwp.lon[idx])  # was not always up to date, so update first

                # Distance to next waypoint where we need to start descent (top of descent) [m]
                descdist = abs(bs.traf.alt[idx] - toalt) / self.steepness  # [m] required length for descent, uses default steepness!
                self.dist2vs[idx] = descdist - xtoalt   # [m] part of that length on this leg

                #print(bs.traf.id[idx],"traf.alt =",bs.traf.alt[idx]/ft,"ft toalt = ",toalt/ft,"ft descdist =",descdist/nm,"nm")
                #print ("d2wp = ",self.dist2wp[idx]/nm,"nm d2vs = ",self.dist2vs[idx]/nm,"nm")
                #print("xtoalt =",xtoalt/nm,"nm descdist =",descdist/nm,"nm")

                # Exceptions: Descend now?
                #print("Active WP:",bs.traf.ap.route[idx].wpname[bs.traf.ap.route[idx].iactwp])
                #print("dist2wp,turndist, dist2vs= ",self.dist2wp[idx],bs.traf.actwp.turndist[idx],self.dist2vs[idx])
                if self.dist2wp[idx] - 1.02*bs.traf.actwp.turndist[idx] < self.dist2vs[idx]:  # Urgent descent, we're late![m]
                    # Descend now using whole remaining distance on leg to reach altitude
                    self.alt[idx] = bs.traf.actwp.nextaltco[idx]  # dial in altitude of next waypoint as calculated
                    t2go = self.dist2wp[idx]/max(0.01,bs.traf.gs[idx])
                    bs.traf.actwp.vs[idx] = (bs.traf.alt[idx]-toalt)/max(0.01,t2go)

                elif xtoalt<descdist: # Not on this leg, no descending is needed at next waypoint
                    # Top of decent needs to be on this leg, as next wp is in descent
                    bs.traf.actwp.vs[idx] = -abs(self.steepness) * (bs.traf.gs[idx] +
                                                                    (bs.traf.gs[idx] < 0.2 * bs.traf.tas[idx]) *
                                                                    bs.traf.tas[idx])

                else:
                    # else still level
                    bs.traf.actwp.vs[idx] = 0.0

            else:

                # We are higher but swtod = False, so there is no ToD descent logic, simply aim at next altco
                steepness_ = (bs.traf.alt[idx]-bs.traf.actwp.nextaltco[idx])/(max(0.01,self.dist2wp[idx]+xtoalt))
                bs.traf.actwp.vs[idx] = -abs(steepness_) * (bs.traf.gs[idx] +
                                                           (bs.traf.gs[idx] < 0.2 * bs.traf.tas[idx]) * bs.traf.tas[
                                                               idx])
                self.dist2vs[idx]      = 99999. #[m] Forces immediate descent as current distance to next wp will be less

                # print("in else swtod for ", bs.traf.id[idx])

        # VNAV climb mode: climb as soon as possible (T/C logic)
        elif bs.traf.alt[idx] < toalt - 9.9 * ft:
            # Stop potential current descent (e.g. due to not making it to previous altco)
            # then stop immediately, as in: do not make it worse.
            if bs.traf.vs[idx] < -0.0001:
                self.vnavvs[idx] = 0.0
                self.alt[idx] = bs.traf.alt[idx]
                if bs.traf.swvnav[idx]:
                    bs.traf.selalt[idx] = bs.traf.alt[idx]

            # Altitude we want to climb to: next alt constraint in our route (could be further down the route)
            bs.traf.actwp.nextaltco[idx] = toalt   # [m]
            bs.traf.actwp.xtoalt[idx]    = xtoalt  # [m] distance to next alt constraint measured from next waypoint
            self.alt[idx]          = bs.traf.actwp.nextaltco[idx]  # dial in altitude of next waypoint as calculated
            self.dist2vs[idx]      = 99999. #[m] Forces immediate climb as current distance to next wp will be less

            t2go = max(0.1, self.dist2wp[idx]+xtoalt) / max(0.01, bs.traf.gs[idx])
            if self.swtoc[idx]:
                steepness_ = self.steepness # default steepness
            else:
                steepness_ = (bs.traf.alt[idx] - bs.traf.actwp.nextaltco[idx]) / (max(0.01, self.dist2wp[idx] + xtoalt))

            bs.traf.actwp.vs[idx]  = np.maximum(steepness_*bs.traf.gs[idx],
                                       (bs.traf.actwp.nextaltco[idx] - bs.traf.alt[idx]) / t2go) # [m/s]
        # Level leg: never start V/S
        else:
            self.dist2vs[idx] = -999.  # [m]

        return

    def setspeedforRTA(self, idx, torta, xtorta):
        #debug print("setspeedforRTA called, torta,xtorta =",torta,xtorta/nm)

        # Calculate required CAS to meet RTA
        # for aircraft nr. idx (scalar)
        if torta < -90. : # -999 signals there is no RTA defined in remainder of route
            return False

        deltime = torta-bs.sim.simt # Remaining time to next RTA [s] in simtime
        if deltime>0: # Still possible?
            gsrta = calcvrta(bs.traf.gs[idx], xtorta,
                             deltime, bs.traf.perf.axmax[idx])

            # Subtract tail wind speed vector
            tailwind = (bs.traf.windnorth[idx]*bs.traf.gsnorth[idx] + bs.traf.windeast[idx]*bs.traf.gseast[idx]) / \
                        bs.traf.gs[idx]

            # Convert to CAS
            rtacas = tas2cas(gsrta-tailwind,bs.traf.alt[idx])

            # Performance limits on speed will be applied in traf.update
            if bs.traf.actwp.spdcon[idx]<0. and bs.traf.swvnavspd[idx]:
                bs.traf.actwp.spd[idx] = rtacas
                #print("setspeedforRTA: xtorta =",xtorta)

            return rtacas
        else:
            return False

    @stack.command(name='ALT')
    def selaltcmd(self, idx: 'acid', alt: 'alt', vspd: 'vspd'=None):
        """ ALT acid, alt, [vspd] 
        
            Select autopilot altitude command."""
        bs.traf.selalt[idx]   = alt
        bs.traf.swvnav[idx]   = False

        # Check for optional VS argument
        if vspd:
            bs.traf.selvs[idx] = vspd
        else:
            if not isinstance(idx, Collection):
                idx = np.array([idx])
            delalt        = alt - bs.traf.alt[idx]
            # Check for VS with opposite sign => use default vs
            # by setting autopilot vs to zero
            oppositevs = np.logical_and(bs.traf.selvs[idx] * delalt < 0., abs(bs.traf.selvs[idx]) > 0.01)

            bs.traf.selvs[idx[oppositevs]] = 0.

    @stack.command(name='VS')
    def selvspdcmd(self, idx: 'acid', vspd:'vspd'):
        """ VS acid,vspd (ft/min)

            Vertical speed command (autopilot) """
        bs.traf.selvs[idx] = vspd #[fpm]
        # bs.traf.vs[idx] = vspd
        bs.traf.swvnav[idx] = False

    @stack.command(name='HDG', aliases=("HEADING", "TURN"))
    def selhdgcmd(self, idx: 'acid', hdg: 'hdg'):  # HDG command
        """ HDG acid,hdg (deg,True or Magnetic)
        
            Autopilot select heading command. """
        if not isinstance(idx, Collection):
            idx = np.array([idx])
        if not isinstance(hdg, Collection):
            hdg = np.array([hdg])
        # If there is wind, compute the corresponding track angle
        if bs.traf.wind.winddim > 0:
            ab50 = bs.traf.alt[idx] > 50.0 * ft
            bel50 = np.logical_not(ab50)
            iab = idx[ab50]
            ibel = idx[bel50]

            tasnorth = bs.traf.tas[iab] * np.cos(np.radians(hdg[ab50]))
            taseast = bs.traf.tas[iab] * np.sin(np.radians(hdg[ab50]))
            vnwnd, vewnd = bs.traf.wind.getdata(bs.traf.lat[iab], bs.traf.lon[iab], bs.traf.alt[iab])
            gsnorth = tasnorth + vnwnd
            gseast = taseast + vewnd
            self.trk[iab] = np.degrees(np.arctan2(gseast, gsnorth))%360.
            self.trk[ibel] = hdg
        else:
            self.trk[idx] = hdg

        bs.traf.swlnav[idx] = False
        # Everything went ok!
        return True

    @stack.command(name='SPD', aliases=("SPEED",))
    def selspdcmd(self, idx: 'acid', casmach: 'spd'):  # SPD command
        """ SPD acid, casmach (= CASkts/Mach) 
        
            Select autopilot speed. """
        # Depending on or position relative to crossover altitude,
        # we will maintain CAS or Mach when altitude changes
        # We will convert values when needed
        bs.traf.selspd[idx] = casmach

        # Used to be: Switch off VNAV: SPD command overrides
        bs.traf.swvnavspd[idx]   = False
        return True

    @stack.command(name='DEST')
    def setdest(self, acidx: 'acid', wpname:'wpt' = None):
        ''' DEST acid, latlon/airport

            Set destination of aircraft, aircraft wil fly to this airport. '''
        if wpname is None:
            return True, 'DEST ' + bs.traf.id[acidx] + ': ' + self.dest[acidx]
        route = self.route[acidx]
        apidx = bs.navdb.getaptidx(wpname)
        if apidx < 0:
            if bs.traf.ap.route[acidx].nwp > 0:
                reflat = bs.traf.ap.route[acidx].wplat[-1]
                reflon = bs.traf.ap.route[acidx].wplon[-1]
            else:
                reflat = bs.traf.lat[acidx]
                reflon = bs.traf.lon[acidx]

            success, posobj = txt2pos(wpname, reflat, reflon)
            if success:
                lat = posobj.lat
                lon = posobj.lon
            else:
                return False, "DEST: Position " + wpname + " not found."

        else:
            lat = bs.navdb.aptlat[apidx]
            lon = bs.navdb.aptlon[apidx]

        self.dest[acidx] = wpname
        iwp = route.addwpt(acidx, self.dest[acidx], route.dest,
                           lat, lon, 0.0, bs.traf.cas[acidx])
        # If only waypoint: activate
        if (iwp == 0) or (self.orig[acidx] != "" and route.nwp == 2):
            bs.traf.actwp.lat[acidx] = route.wplat[iwp]
            bs.traf.actwp.lon[acidx] = route.wplon[iwp]
            bs.traf.actwp.nextaltco[acidx] = route.wpalt[iwp]
            bs.traf.actwp.spd[acidx] = route.wpspd[iwp]

            bs.traf.swlnav[acidx] = True
            bs.traf.swvnav[acidx] = True
            route.iactwp = iwp
            route.direct(acidx, route.wpname[iwp])

        # If not found, say so
        elif iwp < 0:
            return False, ('DEST position'+self.dest[acidx] + " not found.")

    @stack.command(name='ORIG')
    def setorig(self, acidx: 'acid', wpname: 'wpt' = None):
        ''' ORIG acid, latlon/airport

            Set origin of aircraft. '''
        if wpname is None:
            return True, 'ORIG ' + bs.traf.id[acidx] + ': ' + self.orig[acidx]
        route = self.route[acidx]
        apidx = bs.navdb.getaptidx(wpname)
        if apidx < 0:
            if bs.traf.ap.route[acidx].nwp > 0:
                reflat = bs.traf.ap.route[acidx].wplat[-1]
                reflon = bs.traf.ap.route[acidx].wplon[-1]
            else:
                reflat = bs.traf.lat[acidx]
                reflon = bs.traf.lon[acidx]

            success, posobj = txt2pos(wpname, reflat, reflon)
            if success:
                lat = posobj.lat
                lon = posobj.lon
            else:
                return False, ("ORIG: Position " + wpname + " not found.")

        else:
            lat = bs.navdb.aptlat[apidx]
            lon = bs.navdb.aptlon[apidx]

        # Origin: bookkeeping only for now, store in route as origin
        self.orig[acidx] = wpname
        iwp = route.addwpt(acidx, self.orig[acidx], route.orig,
                           lat, lon, 0.0, bs.traf.cas[acidx])
        if iwp < 0:
            return False, (self.orig[acidx] + " not found.")

    @stack.command(name='VNAV')
    def setVNAV(self, idx: 'acid', flag: 'bool'=None):
        """ VNAV acid,[ON/OFF]
        
            Switch on/off VNAV mode, the vertical FMS mode (autopilot) """
        if not isinstance(idx, Collection):
            if idx is None:
                # All aircraft are targeted
                bs.traf.swvnav    = np.array(bs.traf.ntraf * [flag])
                bs.traf.swvnavspd = np.array(bs.traf.ntraf * [flag])
            else:
                # Prepare for the loop
                idx = np.array([idx])

        # Set VNAV for all aircraft in idx array
        output = []
        for i in idx:
            if flag is None:
                msg = bs.traf.id[i] + ": VNAV is " + "ON" if bs.traf.swvnav[i] else "OFF"
                if not bs.traf.swvnavspd[i]:
                    msg += " but VNAVSPD is OFF"
                output.append(bs.traf.id[i] + ": VNAV is " + "ON" if bs.traf.swvnav[i] else "OFF")

            elif flag:
                if not bs.traf.swlnav[i]:
                    return False, (bs.traf.id[i] + ": VNAV ON requires LNAV to be ON")

                route = self.route[i]
                if route.nwp > 0:
                    bs.traf.swvnav[i]    = True
                    bs.traf.swvnavspd[i] = True
                    self.route[i].calcfp()
                    actwpidx = self.route[i].iactwp
                    self.ComputeVNAV(i,self.route[i].wptoalt[actwpidx],self.route[i].wpxtoalt[actwpidx],\
                                     self.route[i].wptorta[actwpidx],self.route[i].wpxtorta[actwpidx])
                    bs.traf.actwp.nextaltco[i] = self.route[i].wptoalt[actwpidx]

                else:
                    return False, ("VNAV " + bs.traf.id[i] + ": no waypoints or destination specified")
            else:
                bs.traf.swvnav[i]    = False
                bs.traf.swvnavspd[i] = False
        if flag == None:
            return True, '\n'.join(output)

    @stack.command(name='LNAV')
    def setLNAV(self, idx: 'acid', flag: 'bool' = None):
        """ LNAV acid,[ON/OFF]

            LNAV (lateral FMS mode) switch for autopilot """
        if not isinstance(idx, Collection):
            if idx is None:
                # All aircraft are targeted
                bs.traf.swlnav = np.array(bs.traf.ntraf * [flag])
            else:
                # Prepare for the loop
                idx = np.array([idx])

        # Set LNAV for all aircraft in idx array
        output = []
        for i in idx:
            if flag is None:
                output.append(bs.traf.id[i] + ": LNAV is " + ("ON" if bs.traf.swlnav[i] else "OFF"))

            elif flag:
                route = self.route[i]
                if route.nwp <= 0:
                    return False, ("LNAV " + bs.traf.id[i] + ": no waypoints or destination specified")
                elif not bs.traf.swlnav[i]:
                    bs.traf.swlnav[i] = True
                    route.direct(i, route.wpname[route.findact(i)])
            else:
                bs.traf.swlnav[i] = False
        if flag is None:
            return True, '\n'.join(output)
        


    @stack.command(name='SWTOC')
    def setswtoc(self, idx: 'acid', flag: 'bool' = None):
        """ SWTOC acid,[ON/OFF]

            Switch ToC logic (=climb early) on/off"""

        if not isinstance(idx, Collection):
            if idx is None:
                # All aircraft are targeted
                self.swtoc = np.array(bs.traf.ntraf * [flag])
            else:
                # Prepare for the loop
                idx = np.array([idx])

        # Set SWTOC for all aircraft in idx array
        output = []
        for i in idx:
            if flag is None:
                output.append(bs.traf.id[i] + ": SWTOC is " + ("ON" if self.swtoc[i] else "OFF"))

            elif flag:
                self.swtoc[i] = True
            else:
                self.swtoc[i] = False
        if flag is None:
            return True, '\n'.join(output)

    @stack.command(name='SWTOD')
    def setswtod(self, idx: 'acid', flag: 'bool' = None):
        """ SWTOD acid,[ON/OFF]

            Switch ToD logic (=climb early) on/off"""
        if not isinstance(idx, Collection):
            if idx is None:
                # All aircraft are targeted
                self.swtod = np.array(bs.traf.ntraf * [flag])
            else:
                # Prepare for the loop
                idx = np.array([idx])

        # Set SWTOD for all aircraft in idx array
        output = []
        for i in idx:
            if flag is None:
                output.append(bs.traf.id[i] + ": SWTOD is " + ("ON" if self.swtoc[i] else "OFF"))

            elif flag:
                self.swtod[i] = True
            else:
                self.swtod[i] = False
        if flag is None:
            return True, '\n'.join(output)


def calcvrta(v0, dx, deltime, trafax):
    # Calculate required target ground speed v1 [m/s]
    # to meet an RTA at this leg
    #
    # Arguments are scalar
    #
    #   v0      = current ground speed [m/s]
    #   dx      = leg distance [m]
    #   deltime = time left till RTA[s]
    #   trafax  = horizontal acceleration [m/s2]

    # Set up variables
    dt = deltime

    # Do we need decelerate or accelerate
    if v0 * dt < dx:
        ax = max(0.01,abs(trafax))
    else:
        ax = -max(0.01,abs(trafax))

    # Solve 2nd order equation for v1 which results from:
    #
    #   dx = 0.5*(v0+v1)*dtacc + v1 * dtconst
    #   dt = trta - tnow = dtacc + dtconst
    #   dtacc = (v1-v0)/ax
    #
    # with unknown dtconst, dtacc, v1
    #
    # -.5/ax * v1**2  +(v0/ax+dt)*v1 -0.5*v0**2 / ax - dx =0

    a = -0.5 / ax
    b = (v0 / ax + dt)
    c = -0.5 * v0 * v0 / ax - dx

    D = b * b - 4. * a * c

    # Possibly two v1 solutions
    vlst = []

    if D >= 0.:
        x1 = (-b - sqrt(D)) / (2. * a)
        x2 = (-b + sqrt(D)) / (2. * a)

        # Check solutions for v1
        for v1 in (x1, x2):
            dtacc = (v1 - v0) / ax
            dtconst = dt - dtacc

            # Physically possible: both dtacc and dtconst >0
            if dtacc >= 0 and dtconst >= 0.:
                vlst.append(v1)

    if len(vlst) == 0:  # Not possible? Maybe borderline, so then simple calculation
        vtarg = dx/dt

    # Just in case both would be valid, take closest to v0
    elif len(vlst) == 2:
        vtarg = vlst[int(abs(vlst[1] - v0) < abs(vlst[0] - v0))]

    # Normal case is one solution
    else:
        vtarg = vlst[0]

    return vtarg

def distaccel(v0,v1,axabs):
    """Calculate distance travelled during acceleration/deceleration
    v0 = start speed, v1 = endspeed, axabs = magnitude of accel/decel
    accel/decel is detemremind by sign of v1-v0
    axabs is acceleration/deceleration of which absolute value will be used
    solve for x: x = vo*t + 1/2*a*t*t    v = v0 + a*t """
    return 0.5*np.abs(v1*v1-v0*v0)/np.maximum(.001,np.abs(axabs))



""" Conditional commands:
KL204 ATSPD 250 KL204 LNAV ON
KL204 ATALT FL100 KL204 SPD 350
"""
import numpy as np
import bluesky as bs
from bluesky import stack
from bluesky.tools.geo import qdrdist

# Enumerated condtion types
alttype, spdtype, postype = 0, 1, 2


class Condition():
    def __init__(self):

        self.ncond = 0  # Number of conditions

        self.id       = []                         # Id of aircraft of condition
        self.condtype = np.array([],dtype=int)     # Condition type (0=alt,1=spd)
        self.target   = np.array([],dtype=float)   # Target value (alt,speed,distance[nm])
        self.lastdif  = np.array([],dtype=float)   # Difference during last update
        self.posdata  = []                         # Data for postype: tuples lat[deg],lon[deg] of ref position
        self.cmd      = []                         # Commands to be issued

    def update(self):
        if self.ncond==0:
            return

        # Update indices based on list of id's
        acidxlst = np.array(bs.traf.id2idx(self.id))
        if len(acidxlst)>0:
            idelcond = sorted(list(np.where(acidxlst<0)[0]))
            for i in idelcond[::-1]:
                del (self.id[i])
                self.condtype = np.delete(self.condtype, i)
                self.target = np.delete(self.target, i)
                self.lastdif = np.delete(self.lastdif, i)
                del self.posdata[i]
                del self.cmd[i]

            self.ncond = len(self.id)
            if self.ncond==0:
                return
            acidxlst = np.array(bs.traf.id2idx(self.id))

        # Check condition types
        actdist = np.ones(self.ncond)*999e9  # Invalid number which never triggers anything is extremely large
        for j in range(self.ncond):
            if self.condtype[j] == postype:
                qdr,dist = qdrdist(bs.traf.lat[acidxlst[j]],bs.traf.lon[acidxlst[j]],self.posdata[j][0],self.posdata[j][1])
                actdist[j] = dist # [nm]

        # Get relevant actual value using index list as index to numpy arrays
        self.actual = (self.condtype == alttype) * bs.traf.alt[acidxlst] + \
                      (self.condtype == spdtype) * bs.traf.cas[acidxlst] + \
                      (self.condtype == postype) * actdist

        # Compare sign of actual difference with sign of last difference
        actdif       = self.target - self.actual

        # Make sorted arrya of indices of true conditions and their conditional commands
        idxtrue      = sorted(list(np.where(actdif*self.lastdif <= 0.0)[0]))# Sign changed
        self.lastdif = actdif
        if idxtrue==None or len(idxtrue)==0:
            return


        # Execute commands found to have true condition
        for i in idxtrue:
            if i>=0:
                stack.stack(self.cmd[i])
                # debug
                # stack.stack(" ECHO Conditional command issued: "+self.cmd[i])

        # Delete executed commands to clean up arrays and lists
        # from highest index to lowest for consistency
        for i in idxtrue[::-1]:
            if i>=0:
                del self.id[i]
                self.condtype = np.delete(self.condtype,i)
                self.target   = np.delete(self.target,i)
                self.lastdif  = np.delete(self.lastdif,i)
                del self.posdata[i]
                del self.cmd[i]

        # Adjust number of conditions
        self.ncond = len(self.id)

        if self.ncond!=len(self.cmd):
            print ("self.ncond=",self.ncond)
            print ("self.cmd=",self.cmd)
            print ("traffic/conditional.py: self.delcondition: invalid condition array size")
        return

    def ataltcmd(self,acidx,targalt,cmdtxt):
        actalt = bs.traf.alt[acidx]
        self.addcondition(acidx, alttype, targalt, actalt, cmdtxt)
        return True

    def atspdcmd(self, acidx, targspd, cmdtxt):
        actspd = bs.traf.tas[acidx]
        self.addcondition(acidx, spdtype, targspd, actspd,cmdtxt)
        return True

    def atdistcmd(self, acidx, lat, lon, targdist, cmdtxt):
        qdr, actdist = qdrdist(bs.traf.lat[acidx], bs.traf.lon[acidx], lat, lon)
        self.addcondition(acidx, postype, targdist, actdist, cmdtxt, (lat,lon))
        return True

    def addcondition(self,acidx, icondtype, target, actual, cmdtxt,latlon=None):
        #print ("addcondition:", acidx, icondtype, target, actual, cmdtxt, latlon)

        # Add condition to arrays
        self.id.append(bs.traf.id[acidx])

        self.condtype = np.append(self.condtype,icondtype)
        self.target   = np.append(self.target,target)
        self.lastdif  = np.append(self.lastdif,target - actual)

        self.posdata.append(latlon)
        self.cmd.append(cmdtxt)

        self.ncond = self.ncond+1
        #print("addcondition: self.ncond",self.ncond)
        return

    def renameac(self,oldid,newid):
        # Continonal commands are stored per id (ac name)
        # When renamed, call this method to update list
        # rename ids in list of ids
        # Call this if RENAME command is implemented
        if self.id.count(old.id) == 0:
            return
        for i in range(len(self.id)):
            if self.id[i] == oldid:
                self.id[i] = newid
        return








import socket
import binascii
from os import urandom
from typing import Union
from enum import Enum, auto


# Message headers (first byte): (un)subscribe
MSG_SUBSCRIBE = 1
MSG_UNSUBSCRIBE = 0
# Message headers (second byte): group identifiers
GROUPID_DEFAULT = 0
GROUPID_CLIENT = ord('C')
GROUPID_SIM = ord('S')
GROUPID_NOGROUP = ord('N')
# Connection identifier string length
IDLEN = 5

class ActionType(Enum):
    ''' Shared state action types. 
    
        An incoming shared state update can be of the following types:
        Append: An item is appended to the state
        Extend: Two or more items are appended to the state
        Delete: One or more items are deleted from the state
        Update: One or more items within the state are updated
        Replace: The full state object is replaced
        Reset: The entire object is reset to its (empty) default
        ActChange: A new active remote is selected
    '''
    Append = b'A'
    Extend = b'E'
    Delete = b'D'
    Update = b'U'
    Replace = b'R'
    Reset = b'X'
    ActChange = b'C'
    NoAction = b''

    @classmethod
    def isaction(cls, data):
        ''' Returns True if passed data is an ActionType '''
        return any([data == a.value for a in cls])


class MessageType(Enum):
    ''' BlueSky network message type indicator. '''
    Unknown = auto()
    Regular = auto()
    SharedState = auto()


def genid(groupid: Union[str, bytes, int]='', idlen=IDLEN, seqidx=1):
    ''' Generate a binary identifier string 
    
        The identifier string consists of a group identifier of idlen-1 bytes,
        and ends with a sequence number that is indicated with curidx.

        Arguments:
        - groupid: The group identifier of the generated id can be part of a
          larger group. A group id passed to the function is extended with random
          bytes up to a length of idlen-1. Valid values in each position are all
          possible byte values except the wildcard character '*', which is reserved
          as wildcard padding.

        - idlen: The length in bytes of the generated identifier string

        - seqidx: The value of the sequence index part of this identifier.
    '''
    groupid = asbytestr(groupid)
    if len(groupid) >= idlen:
        return groupid
    elif len(groupid) < idlen - 1:
        groupid += urandom(idlen - 1 - len(groupid)).replace(b'*', b'_')
    return groupid + seqidx2id(seqidx)


def seqid2idx(seqid):
    ''' Transform a bytestring sequence id to a numeric sequence index.
        Returns -1 if this is not a valid sequence id.
    '''
    ret = (seqid if isinstance(seqid, int) else ord(seqid)) - 128
    return max(-1, ret)


def seqidx2id(seqidx):
    ''' Transform a numeric sequence index to a bytestring sequence id '''
    return chr(128 + seqidx).encode('charmap')


def asbytestr(val:int|str|bytes) -> bytes:
    return chr(val).encode('charmap') if isinstance(val, int) else \
              val.encode('charmap') if isinstance(val, str) else \
              val

def bin2hex(bstr):
    return binascii.hexlify(bstr).decode()


def hex2bin(hstr):
    return binascii.unhexlify(hstr)


def get_ownip():
    ''' Try to determine the IP address of this machine. '''
    try:
        local_addrs = socket.gethostbyname_ex(socket.gethostname())[-1]

        for addr in local_addrs:
            if not addr.startswith('127'):
                return addr
    except:
        pass
    return '127.0.0.1'



''' BlueSky colour palette loader. '''
import bluesky as bs
bs.settings.set_variable_defaults(colour_palette='bluesky-default', gfx_path='graphics')

def init():
    # Load the palette file selected in settings
    pfile = bs.resource(bs.settings.gfx_path) / 'palettes' / bs.settings.colour_palette
    if pfile.is_file():
        print('Loading palette ' + bs.settings.colour_palette)
        exec(compile(open(pfile).read(), pfile, 'exec'), globals())
        return True
    else:
        print('Palette file not found ' + pfile)
        return False


def set_default_colours(**kwargs):
    ''' Register a default value for a colour. Use this functionality in the source file
        where you intend to use those colours so that defaults are always available.

        Example:
            from bluesky.ui import palette
            palette.set_default_colours(mycolor1=(255, 0, 0), mycolor2=(0, 0, 0))

            This will make settings.mycolor1 and settings.mycolor2 available,
            with the provided default values.'''
    for key, value in kwargs.items():
        if key not in globals():
            globals()[key] = value

initialized = init()



''' Tile texture manager for BlueSky Qt/OpenGL gui. '''
import traceback
import math
import weakref
from collections import OrderedDict
from urllib.request import urlopen
from urllib.error import URLError
import numpy as np
import gzip

from PyQt6.QtCore import Qt
from PyQt6.QtCore import QObject, QRunnable, QThreadPool, pyqtSignal, pyqtSlot
from PyQt6.QtGui import QImage, qRgba

import bluesky as bs
from bluesky.network.subscriber import subscribe
from bluesky.ui.qtgl import glhelpers as glh


bs.settings.set_variable_defaults(
    tile_array_size=100,
    max_download_workers=2,
    max_tile_zoom=18,
    tile_sources={
        'opentopomap': {
            'source': ['https://a.tile.opentopomap.org/{zoom}/{x}/{y}.png',
                       'https://b.tile.opentopomap.org/{zoom}/{x}/{y}.png',
                       'https://c.tile.opentopomap.org/{zoom}/{x}/{y}.png'],
            'max_download_workers': 2,
            'max_tile_zoom': 17,
            'license': 'map data: © OpenStreetMap contributors, SRTM | map style: © OpenTopoMap.org (CC-BY-SA)'},
        'cartodb': {
            'source': ['https://cartodb-basemaps-b.global.ssl.fastly.net/light_nolabels/{zoom}/{x}/{y}.png'],
            'max_tile_zoom': 20,
            'license': 'CartoDB Grey and white, no labels'
        },
        'nasa': {
            'source': ['https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{zoom}/{y}/{x}.jpg'],
            'max_tile_zoom': 13,
            'license': 'Satellite images from NASA via ESRI'
        }
    })


class Tile:
    ''' Wrapper object for tile data and properties. '''

    def __init__(self, source, zoom, tilex, tiley, idxx, idxy):
        super().__init__()
        self.source = source
        self.zoom = zoom
        self.tilex = tilex
        self.tiley = tiley
        self.idxx = idxx
        self.idxy = idxy
        self.ext = source[source.rfind('.'):]
        self.ext = '.' + bs.settings.tile_sources[source]['source'][0].split('.')[-1]

        self.image = None
        # For the image data, check cache path first
        fpath = bs.resource(bs.settings.cache_path) / source / str(zoom) / str(tilex)
        fname = fpath / f'{tiley}{self.ext}'
        if fname.exists():
            self.image = QImage(fname.as_posix()).convertToFormat(QImage.Format.Format_ARGB32)
        else:
            # Make sure cache directory exists
            fpath.mkdir(parents=True, exist_ok=True)
            for url in bs.settings.tile_sources[source]['source']:
                try:
                    url_request = urlopen(url.format(
                        zoom=zoom, x=tilex, y=tiley))
                    
                    if url_request.status == 204:
                        # if no content load a blank tile
                        self.image = QImage(256, 256, QImage.Format.Format_ARGB32)
                        self.image.fill(qRgba(255,255,255,0))

                    else:
                        data = url_request.read()

                        if url_request.headers['Content-Encoding'] == 'gzip':
                            # There is a chance that data may come as a gzip so decompress
                            data = gzip.decompress(data)
                                        
                        self.image = QImage.fromData(
                            data).convertToFormat(QImage.Format.Format_ARGB32)

                        with open(fname, 'wb') as fout:
                            fout.write(data)
                    break
                except URLError as e:
                    print(f'Error loading {url.format(zoom=zoom, x=tilex, y=tiley)}:')
                    print(traceback.format_exc())


class TileLoader(QRunnable):
    ''' Thread worker to load tiles in the background. '''
    class Signals(QObject):
        ''' Container class for worker signals. '''
        finished = pyqtSignal(Tile)

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.signals = TileLoader.Signals()

    def run(self):
        ''' Function to execute in the worker thread. '''
        tile = Tile(*self.args, **self.kwargs)
        if tile.image is not None:
            self.signals.finished.emit(tile)


class TiledTextureMeta(type(glh.Texture)):
    ''' TileTexture meta class that stores weak references to created textures
        so that they can be used in multiple objects, but are deleted when the
        last reference to them is removed. '''
    tiletextures = weakref.WeakValueDictionary()
    def __call__(cls, *args, **kwargs):
        name = kwargs.get('tilesource', 'opentopomap')
        tex = cls.tiletextures.get(name)
        if tex is None:
            tex = super().__call__(*args, **kwargs)
            cls.tiletextures[name] = tex
        return tex


class TiledTexture(glh.Texture, metaclass=TiledTextureMeta):
    ''' Tiled texture implementation for the BlueSky GL gui. '''
    class SlotHolder(QObject):
        ''' Wrapper class for Qt slot, which can only be owned by a
            QObject-derived parent. We need slots to allow signal receiver
            to be executed in the receiving (main) thread. '''
        def __init__(self, callback):
            super().__init__()
            self.cb = callback

        @pyqtSlot(Tile)
        def slot(self, *args, **kwargs):
            self.cb(*args, **kwargs)

    def __init__(self, glsurface, tilesource='opentopomap'):
        super().__init__(target=glh.Texture.Target.Target2DArray)
        self.threadpool = QThreadPool()
        tileinfo = bs.settings.tile_sources.get(tilesource)
        if not tileinfo:
            raise KeyError(f'Tile source {tilesource} not found!')
        max_dl = tileinfo.get('max_download_workers', bs.settings.max_download_workers)
        self.maxzoom = tileinfo.get('max_tile_zoom', bs.settings.max_tile_zoom)
        self.threadpool.setMaxThreadCount(min(bs.settings.max_download_workers, max_dl))
        self.tileslot = TiledTexture.SlotHolder(self.load_tile)
        self.tilesource = tilesource
        self.tilesize = (256, 256)
        self.curtileext = (0, 0, 0, 0)
        self.curtilezoom = 1
        self.curtiles = OrderedDict()
        self.fullscreen = False
        self.offsetscale = np.array([0, 0, 1], dtype=np.float32)
        self.bbox = list()
        self.glsurface = glsurface
        self.indextexture = glh.Texture(target=glh.Texture.Target.Target2D)
        self.indexsampler_loc = 0
        self.arraysampler_loc = 0
        subscribe('PANZOOM').connect(self.panzoom)

    def add_bounding_box(self, lat0, lon0, lat1, lon1):
        ''' Add the bounding box of a textured shape.

            These bounding boxes are used to determine if tiles need to be
            downloaded. '''
        if abs(lat1 - lat0) >= 178 and abs(lon1 - lon0) >= 358:
            self.fullscreen = True
        else:
            self.bbox.append((lat0, lon0, lat1, lon1))

    def create(self):
        ''' Create this texture in GPU memory. '''

        if self.isCreated():
            return
        super().create()

        # Fetch a temporary tile image to get dimensions
        tmptile = Tile(self.tilesource, 1, 1, 1, 0, 0)
        img = tmptile.image
        self.setFormat(glh.Texture.TextureFormat.RGBA8_UNorm)
        self.tilesize = (img.width(), img.height())
        self.setSize(img.width(), img.height())
        self.setLayers(bs.settings.tile_array_size)
        super().bind()
        self.allocateStorage()
        self.setWrapMode(glh.Texture.CoordinateDirection.DirectionS,
                         glh.Texture.WrapMode.ClampToBorder)
        self.setWrapMode(glh.Texture.CoordinateDirection.DirectionT,
                         glh.Texture.WrapMode.ClampToBorder)
        self.setMinMagFilters(glh.Texture.Filter.Linear, glh.Texture.Filter.Linear)

        # Initialize index texture
        # RG = texcoord offset, B = zoom factor, A = array index
        itexw = int(np.sqrt(bs.settings.tile_array_size) * 4 / 3 + 10)
        itexh = int(np.sqrt(bs.settings.tile_array_size) * 3 / 4 + 10)
        self.indextexture.create()
        self.indextexture.setFormat(glh.Texture.TextureFormat.RGBA32I)
        self.indextexture.setSize(itexw, itexh)
        self.indextexture.bind(1)
        # self.indextexture.allocateStorage(glh.Texture.RGBA_Integer, glh.Texture.Int32)

        idxdata = np.array(itexw * itexh *
                           [(0, 0, 0, -1)], dtype=np.int32)
        target_value = glh.Texture.Target.Target2D.value
        text_rgba_value = glh.Texture.TextureFormat.RGBA32I.value
        pixel_rgba_value = glh.Texture.PixelFormat.RGBA_Integer.value
        pixel_type_value = glh.Texture.PixelType.Int32.value

        glh.gl.glTexImage2D_alt(target_value, 0, text_rgba_value,
                                itexw, itexh, 0, pixel_rgba_value,
                                pixel_type_value, idxdata.tobytes())

        self.indextexture.setWrapMode(glh.Texture.CoordinateDirection.DirectionS,
                                      glh.Texture.WrapMode.ClampToBorder)
        self.indextexture.setWrapMode(glh.Texture.CoordinateDirection.DirectionT,
                                      glh.Texture.WrapMode.ClampToBorder)
        self.indextexture.setMinMagFilters(glh.Texture.Filter.Nearest, glh.Texture.Filter.Nearest)

        shader = glh.ShaderSet.get_shader('tiled')
        self.indexsampler_loc = shader.uniformLocation('tile_index')
        self.arraysampler_loc = shader.uniformLocation('tile_texture')

    def bind(self, unit=0):
        ''' Bind this texture for drawing. '''
        # Set sampler locations
        glh.ShaderProgram.bound_shader.setUniformValue(self.indexsampler_loc, 2)
        glh.ShaderProgram.bound_shader.setUniformValue(self.arraysampler_loc, 4)
        # Bind index texture to texture unit 0
        self.indextexture.bind(2)
        # Bind tile array texture to texture unit 1
        super().bind(4)

    def panzoom(self, pzdata=None, finished=True):
        ''' Update textures whenever pan/zoom changes. 
            
            Arguments:
            - finished: False when still in the process of panning/zooming.
        '''
        # Check if textures need to be updated
        viewport = self.glsurface.viewportlatlon()
        surfwidth_px = self.glsurface.width()

        # First determine floating-point, hypothetical values
        # to calculate the required tile zoom level
        # floating-point number of tiles that fit in the width of the view
        ntiles_hor = surfwidth_px / self.tilesize[0]
        # Estimated width in longitude of one tile
        est_tilewidth = abs(viewport[3] - viewport[1]) / ntiles_hor

        self.curtilezoom = tilezoom(est_tilewidth, self.maxzoom)
        # With the tile zoom level get the required number of tiles
        # Top-left and bottom-right tiles:
        x0, y0 = latlon2tilenum(*viewport[:2], self.curtilezoom)
        x1, y1 = latlon2tilenum(*viewport[2:], self.curtilezoom)
        nx = abs(x1 - x0) + 1
        ny = abs(y1 - y0) + 1
        self.curtileext = (x0, y0, x1, y1)

        # Calculate the offset of the top-left tile w.r.t. the screen top-left corner
        tile0_topleft = np.array(tilenum2latlon(x0, y0, self.curtilezoom))
        tile0_bottomright = np.array(tilenum2latlon(x0 + 1, y0 + 1, self.curtilezoom))
        tilesize_latlon0 = np.abs(tile0_bottomright - tile0_topleft)
        offset_latlon0 = viewport[:2] - tile0_topleft
        tex_y0, tex_x0 = np.abs(offset_latlon0 / tilesize_latlon0)

        # Calculate the offset of the bottom-right tile w.r.t. the screen bottom right corner
        tile1_topleft = np.array(tilenum2latlon(x1, y1, self.curtilezoom))
        tile1_bottomright = np.array(tilenum2latlon(x1 + 1, y1 + 1, self.curtilezoom))
        tilesize_latlon1 = np.abs(tile1_bottomright - tile1_topleft)
        offset_latlon1 = viewport[2:] - tile1_topleft
        tex_y1, tex_x1 = np.abs(offset_latlon1 / tilesize_latlon1) + [ny - 1, nx - 1]

        # Store global offset and scale for shader uniform
        self.offsetscale = np.array(
            [tex_x0, tex_y0, tex_x1 - tex_x0, tex_y1 - tex_y0], dtype=np.float32)
        # Determine required tiles
        index_tex = []

        curtiles = OrderedDict()
        curtiles_difzoom = OrderedDict()
        for j, y in enumerate(range(y0, y1 + 1)):
            for i, x in enumerate(range(x0, x1 + 1)):
                # Find tile index if already loaded
                idx = self.curtiles.pop((x, y, self.curtilezoom), None)
                if idx is not None:
                    # correct zoom, so dx,dy=0, zoomfac=1
                    index_tex.extend((0, 0, 1, idx))
                    curtiles[(x, y, self.curtilezoom)] = idx
                else:
                    if finished:
                        # Tile not loaded yet, fetch in the background
                        task = TileLoader(self.tilesource, self.curtilezoom, x, y, i, j)
                        task.signals.finished.connect(self.tileslot.slot, Qt.ConnectionType.QueuedConnection)
                        self.threadpool.start(task)

                    # In the mean time, check if more zoomed-out tiles are loaded that can be used
                    for z in range(self.curtilezoom - 1, max(2, self.curtilezoom - 5), -1):
                        zx, zy, dx, dy = zoomout_tilenum(x, y, z - self.curtilezoom)
                        idx = self.curtiles.pop((zx, zy, z), None)
                        if idx is not None:
                            curtiles_difzoom[(zx, zy, z)] = idx
                            zoomfac = int(2 ** (self.curtilezoom - z))
                            dxi = int(round(dx * zoomfac))
                            dyi = int(round(dy * zoomfac))
                            # offset zoom, so possible offset dxi, dyi
                            index_tex.extend((dxi, dyi, zoomfac, idx))
                            break
                    else:
                        # No useful tile found
                        index_tex.extend((0, 0, 0, -1))
        # Update curtiles ordered dict
        curtiles.update(curtiles_difzoom)
        curtiles.update(self.curtiles)
        self.curtiles = curtiles

        data = np.array(index_tex, dtype=np.int32)
        self.glsurface.makeCurrent()
        self.indextexture.bind(2)

        target_value = glh.Texture.Target.Target2D.value
        pixel_rgba_value = glh.Texture.PixelFormat.RGBA_Integer.value
        pixel_type_value = glh.Texture.PixelType.Int32.value

        glh.gl.glTexSubImage2D_alt(target_value, 0, 0, 0, nx, ny,
                                  pixel_rgba_value, pixel_type_value, data.tobytes())

    def load_tile(self, tile):
        ''' Send loaded image data to GPU texture array.

            This function is called on callback from the
            asynchronous image load function.
        '''
        # First check whether tile is still relevant
        # If there's a lot of panning/zooming, sometimes tiles are loaded that
        # become obsolete before they are even downloaded.
        if not ((self.curtilezoom == tile.zoom) and
                (self.curtileext[0] <= tile.tilex <= self.curtileext[2]) and
                (self.curtileext[1] <= tile.tiley <= self.curtileext[3])):
            return

        # Make sure our GL context is current
        self.glsurface.makeCurrent()

        # Check if tile is already loaded. This can happen here with multiple
        # pans/zooms shortly after each other
        layer = self.curtiles.get((tile.tilex, tile.tiley, tile.zoom), None)
        if layer is None:
            # This tile is not loaded yet. Select layer to upload it to
            layer = len(self.curtiles)
            if layer >= bs.settings.tile_array_size:
                # we're exceeding the size of the GL texture array.
                # Replace the least-recent tile
                _, layer = self.curtiles.popitem()

            self.curtiles[(tile.tilex, tile.tiley, tile.zoom)] = layer
            # Upload tile to texture array
            super().bind(4)
            self.setLayerData(layer, tile.image)
            self.release()

        # Update the ordering of the tile dict: the new tile should be on top
        self.curtiles.move_to_end(
            (tile.tilex, tile.tiley, tile.zoom), last=False)

        # Update index texture
        idxdata = np.array([0, 0, 1, layer], dtype=np.int32)
        self.indextexture.bind(2)

        target_value = glh.Texture.Target.Target2D.value
        pixel_rgba_value = glh.Texture.PixelFormat.RGBA_Integer.value
        pixel_type_value = glh.Texture.PixelType.Int32.value


        glh.gl.glTexSubImage2D_alt(target_value, 0, tile.idxx, tile.idxy,
                                   1, 1, pixel_rgba_value,
                                   pixel_type_value, idxdata.tobytes())

        self.indextexture.release()


def latlon2tilenum(lat_deg, lon_deg, zoom):
    ''' Generate tile x/y numbers for a given lat/lon/zoom. '''
    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = int((lon_deg + 180.0) / 360.0 * n)
    ytile = int((1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n)
    return (xtile, ytile)


def tilenum2latlon(xtile, ytile, zoom):
    ''' Generate lat/lon coordinates for the top-left corner of a tile, given
        the x/y tilenumbers and zoom factor. '''
    n = 2.0 ** zoom
    lon_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return (lat_deg, lon_deg)


def tilewidth(zoom):
    ''' Calculate the width of a tile in degrees longitude,
        given a zoom factor. '''
    n = 2 ** zoom
    dlon_deg = 1.0 / n * 360.0
    return dlon_deg


def tilezoom(dlon, maxzoom=18):
    ''' Calculate a zoom factor, given a (hypothetical) width of a tile in
        degrees longitude. '''
    n = 1.0 / dlon * 360.0
    zoom = math.log2(n)
    return min(maxzoom, round(zoom))


def zoomout_tilenum(xtile_in, ytile_in, delta_zoom):
    ''' Calculate corresponding tile x/y number for the overlapping tile
        that is 'delta_zoom' steps zoomed out. '''
    zoomfac = 2 ** delta_zoom
    xtilef = xtile_in * zoomfac
    ytilef = ytile_in * zoomfac
    xtile = int(xtilef)
    dx = xtilef - xtile
    ytile = int(ytilef)
    dy = ytilef - ytile
    return xtile, ytile, dx, dy



import inspect

import bluesky as bs
from bluesky.core.funcobject import FuncObject
from bluesky.core import signal
import bluesky.network.sharedstate as ss
import bluesky.network.context as ctx
from bluesky.network.common import GROUPID_DEFAULT, ActionType, MessageType


#TODO:
# trigger voor actnode changed?


def subscriber(func=None, *, topic='', broadcast=True, actonly=False, raw=False, from_group=GROUPID_DEFAULT, to_group=''):
    ''' BlueSky network subscription decorator.

        Functions decorated with this decorator will be called whenever data
        with the specified topic is received.

        Arguments:
        - topic:      The topic to subscribe to for this function. Function name
                      is used when no topic is specified
        - broadcast:  Whether to subscribe to broadcast-to-(all/group) topic.
                      When set to False, to_group and from_group are ignored,
                      and only messages sent directly to this client/node are
                      received.
        - raw:        Set to true for SharedState messages if you want to
                      receive the original unprocessed message
        - from_group: Subscribe to data from a specific sender(-group) (optional)
                      By default, broadcast subscriptions filter so that only
                      sim->client and client->sim messages are received (to avoid
                      receiving our own broadcasts). If you don't want this,
                      set from_group='*'.
        - to_group:   Subscribe to data sent to a specific group (optional)
        - actonly:    Only receive this data for the active node (client only)
    '''
    def deco(func):
        # Unwrap wrapped functions, static and class methods
        ifunc = inspect.unwrap(func, stop=lambda f:not isinstance(f, (staticmethod, classmethod)))
        
        # Create the subscription object. Network subscriptions will be made as
        # soon as the network connection is available
        Subscription(topic.upper() or ifunc.__name__.upper(), broadcast, actonly, from_group, to_group).connect(ifunc, raw)

        # Construct the subscription object, but return the original function
        return func
    # Allow both @subscriber and @subscriber(args)
    return deco if func is None else deco(func)


def subscribe(topic, *, broadcast=True, actonly=False, from_group=GROUPID_DEFAULT, to_group='') -> 'Subscription':
    ''' Subscribe to a network topic without passing a function callback. 
    
        Arguments:
        - topic: The topic to subscribe to for this function.
        - broadcast:  Whether to subscribe to broadcast-to-(all/group) topic.
                      When set to False, to_group and from_group are ignored,
                      and only messages sent directly to this client/node are
                      received.
        - from_group: Subscribe to data from a specific sender(-group) (optional)
                      By default, broadcast subscriptions filter so that only
                      sim->client and client->sim messages are received (to avoid
                      receiving our own broadcasts). If you don't want this,
                      set from_group='*'.
        - to_group:   Subscribe to data sent to a specific group (optional)
        - actonly:    Only receive this data for the active node (client only)

        Returns:
        - The subscription object for this topic
    '''
    # Create the subscription object. Network subscriptions will be made as
    # soon as the network connection is available
    return Subscription(topic.upper(), broadcast, actonly, from_group, to_group)


class SubscriptionFactory(signal.SignalFactory):
    # Individually keep track of all network subscriptions
    subscriptions: dict[str, 'Subscription'] = dict()

    def __call__(cls, topic, broadcast=True, actonly=None, from_group=GROUPID_DEFAULT, to_group=''):
        ''' Factory function for Subscription construction. 
        
            Arguments:
            - topic:      The topic to subscribe to for this function. Function name
                          is used when no topic is specified
            - broadcast:  Whether to subscribe to broadcast-to-(all/group) topic.
                          When set to False, to_group and from_group are ignored,
                          and only messages sent directly to this client/node are
                          received.
            - from_group: Subscribe to data from a specific sender(-group) (optional)
                          By default, broadcast subscriptions filter so that only
                          sim->client and client->sim messages are received (to avoid
                          receiving our own broadcasts). If you don't want this,
                          set from_group='*'.
            - to_group:   Subscribe to data sent to a specific group (optional)
            - actonly:    Only receive this data for the active node (client only)
        '''
        # # Convert name to string if necessary
        if isinstance(topic, bytes):
            topic = topic.decode()
        # Get subscription object if it already exists
        sub = SubscriptionFactory.subscriptions.get(topic)
        if sub is None:
            sub = super().__call__(topic)

            # Check if type is correct. It could be that Subscription was previously
            # initialised as plain Signal
            if not isinstance(sub, Subscription):
                if issubclass(Subscription, type(sub)):
                    # Signal object instance needs to stay intact, so instead change type and reinitialise
                    sub.__class__ = Subscription
                    sub.subs = set()
                    sub.requested = set()
                    sub.actonly = False
                    sub.msg_type = MessageType.Unknown
                    sub.deferred_subs = []
                    signal.Signal.connect(sub, sub._detect_type)
                else:
                    raise TypeError(f'Trying to connect network subscription to signal with incompatible type {type(sub).__name__}')

            # Store subscription
            SubscriptionFactory.subscriptions[topic] = sub

        # Keep track of all broadcast topic orig/dest combinations
        if actonly is not None:
            sub.actonly = actonly
        if broadcast or sub.actonly:
            sub.subscribe(from_group, to_group)

        return sub


class Subscription(signal.Signal, metaclass=SubscriptionFactory):
    ''' Network subscription class.
    
        Objects of this type are used to store subscription details, and to 
        keep track of local topic subscribers
    '''
    def __init__(self, topic, broadcast=True, actonly=False, from_group=GROUPID_DEFAULT, to_group=''):
        super().__init__(topic)
        self.subs = set()
        self.requested = set()
        self.actonly = actonly
        self.deferred_subs = []
        if ss.is_sharedstate(topic):
            # We already know this is a sharedstate topic. subscribe as such
            self.msg_type = MessageType.SharedState
            super().connect(ss.on_sharedstate_received)
        else:
            # Start out uninitialised so we can detect whether incoming data is
            # a shared state or a regular message
            self.msg_type = MessageType.Unknown
            super().connect(self._detect_type)

    def _detect_type(self, *args, **kwargs):
        # First disconnect this function, it is only needed once
        super().disconnect(self._detect_type)
        # This function responds to the first incoming message for topic
        # Detect whether it is a sharedstate message
        if args and ActionType.isaction(args[0]):
            # This is a sharedstate message
            self.msg_type = MessageType.SharedState
            # In this case, all (non-raw) subscribers will be configured
            # as sharedstate subscribers
            sig = ss.addtopic(self.topic)
            while self.deferred_subs:
                sig.connect(self.deferred_subs.pop())

            # Finally send the sharedstate message on to the subscribers,
            # and subscribe the sharedstate processing function to this topic
            super().connect(ss.on_sharedstate_received)
            ss.on_sharedstate_received(*args, **kwargs)

        else:
            self.msg_type = MessageType.Regular
            while self.deferred_subs:
                cb = self.deferred_subs.pop()
                cb(*args, **kwargs)
                super().connect(cb)


    @property
    def active(self):
        ''' Returns True if this Subscription has activated network subscriptions. '''
        return bool(self.subs)

    def connect(self, func, raw=False):
        ''' Connect a callback function to incoming data on this
            subscription topic.
            
            Arguments:
            - func: The callback function
            - raw:  Set this to True for SharedState messages if you want to
                    receive the original unprocessed message
        '''
        self.subscribe_all()
        if raw or self.msg_type == MessageType.Regular:
            super().connect(func)
        elif self.msg_type == MessageType.Unknown:
            self.deferred_subs.append(FuncObject(func))
        else:
            signal.Signal(f'state-changed.{self.topic.lower()}').connect(func)

    def subscribe(self, from_group=GROUPID_DEFAULT, to_group=''):
        if (from_group, to_group) not in self.subs:
            if bs.net is not None:
                self.subs.add((from_group, to_group))
                if self.actonly:
                    bs.net._subscribe(self.topic, from_group, to_group, self.actonly)
                else:
                    bs.net._subscribe(self.topic, from_group, to_group)
            else:
                self.requested.add((from_group, to_group))

    def subscribe_all(self):
        if bs.net is not None:
            while self.requested:
                self.subscribe(*self.requested.pop())

    def unsubscribe(self, from_group=GROUPID_DEFAULT, to_group=''):
        if (from_group or to_group is not None):
            if (from_group, to_group) in self.subs:
                self.subs.discard((from_group, to_group))
                if bs.net is not None:
                    bs.net._unsubscribe(self.topic, from_group, to_group)
        else:
            # Unsubscribe all
            self.requested.clear()
            if bs.net is not None:
                while self.subs:
                    bs.net._unsubscribe(self.topic, *self.subs.pop())


@subscriber
def reset(*args):
    ''' Process incoming RESET events. '''
    # Clear state data to defaults for this simulation node
    ss.reset(ctx.sender_id)


@signal.subscriber(topic='node-added')
def on_node_added(node_id):
    ''' When a new node is announced, request the initial/current state of all 
        subscribed shared states.
    '''
    topics = [topic for topic, sub in SubscriptionFactory.subscriptions.items() 
              if sub.msg_type in (MessageType.Unknown, MessageType.SharedState)]
    bs.net.send('REQUEST', topics, to_group=node_id)



""" Node encapsulates the sim process, and manages process I/O. """
import os
import bluesky as bs
from bluesky.core.walltime import Timer


class Node:
    def __init__(self, *args):
        self.node_id = b'\x00' + os.urandom(4)
        self.server_id = b''
        self.act_id = None
        self.running = True

    def update(self):
        ''' Update timers. '''
        # Process timers
        Timer.update_timers()

    def stop(self):
        ''' Stack stop/quit command. '''
        # On a stack quit command, detached simulation just quits.
        self.quit()

    def quit(self):
        ''' Quit the simulation process. '''
        self.running = False

    def run(self):
        ''' Start the main loop of this node. '''
        while self.running:
            # Perform a simulation step
            self.update()
            bs.sim.step()

    def connect(self):
        ''' Connect node to the BlueSky server. This does nothing in detached mode. '''
        pass

    def addnodes(self, count=1):
        pass

    def send(self, topic, data='', to_group=b''):
        pass

    def subscribe(self, topic, from_id=b'', to_group=b''):
        pass

    def close(self):
        pass


''' Stack argument parsers. '''
import inspect
import re
from types import UnionType, NoneType
import typing
from matplotlib import colors
from bluesky.tools.misc import txt2bool, txt2lat, txt2lon, txt2alt, txt2tim, \
    txt2hdg, txt2vs, txt2spd
from bluesky.tools.position import Position, islat
import bluesky as bs


# Regular expression for argument parser
# Reading the regular expression:
# [\'"]?             : skip potential opening quote
# (?<=[\'"])[^\'"]+  : look behind for a leading quote, and if so, parse everything until closing quote
# (?<![\'"])[^\s,]+  : look behind for not a leading quote, then parse until first whitespace or comma
# [\'"]?\s*,?\s*     : skip potential closing quote, whitespace, and a potential single comma
re_getarg = re.compile(
    r'\s*[\'"]?((?<=[\'"])[^\'"]*|(?<![\'"])[^\s,]*)[\'"]?\s*,?\s*(.*)')
# re_getarg = re.compile(r'[\'"]?((?<=[\'"])[^\'"]+|(?<![\'"])[^\s,]+)[\'"]?\s*,?\s*')


def getnextarg(cmdstring):
    ''' Return first argument and remainder of command string from cmdstring. '''
    return re_getarg.match(cmdstring).groups()


class Parameter:
    ''' Wrapper class for stack function parameters. '''
    def __init__(self, param, annotation='', isopt=None):
        self.name = param.name
        self.default = param.default
        self.optional = (self.hasdefault() or param.kind == param.VAR_POSITIONAL) if isopt is None else isopt
        self.gobble = param.kind == param.VAR_POSITIONAL and not annotation
        self.annotation = annotation or param.annotation

        # Make list of parsers
        if self.annotation is inspect._empty:
            # Without annotation the argument is passed on unchanged as string
            # (i.e., the 'word' argument type)
            self.parsers = [Parser(str)]
            self.annotation = 'word'
        elif isinstance(self.annotation, str):
            # If the annotation is a string we get our parsers from the argparsers dict
            pfuns = [argparsers.get(a) for a in self.annotation.split('/')]
            self.parsers = [p for p in pfuns if p is not None]
        elif isinstance(param.annotation, type) and issubclass(param.annotation, Parser):
            # If the paramter annotation is a class derived from Parser
            self.parsers = [self.annotation()]

        # All other annotation types are expected to have default behaviour
        # and are wrapped in Parser
        elif isinstance(self.annotation, UnionType):
            self.parsers = []
            for tp in typing.get_args(self.annotation):
                if tp is NoneType:
                    self.optional = True
                else:
                    self.parsers.append(Parser(tp))
        else:
            self.parsers = [Parser(self.annotation)]

        # This parameter is not valid if it has no parsers, or is keyword-only.
        # In those cases it can be skipped from the list of parameters when
        # processing a stack command line.
        self.valid = bool(self.parsers) and self.canwrap(param)

    def __call__(self, argstring):
        # First check if argument is omitted and default value is needed
        if not argstring or argstring[0] == ',':
            _, argstring = re_getarg.match(argstring).groups()
            if self.hasdefault():
                return self.default, argstring
            if self.optional:
                return (None, argstring) if argstring else ('',)
            raise ArgumentError(f'Missing argument {self.name}')
        # Try available parsers
        error = ''
        for parser in self.parsers:
            try:
                return parser.parse(argstring)
            except (ValueError, ArgumentError) as e:
                error += ('\n' + e.args[0])

        # If all fail, raise error
        raise ArgumentError(error)

    def __str__(self):
        return f'{self.name}:{self.annotation}'

    def __bool__(self):
        return self.valid

    def size(self):
        ''' Returns the (maximum) number of return variables when parsing this
            parameter. '''
        return max((p.size for p in self.parsers))

    def hasdefault(self):
        ''' Returns True if this parameter has a default value. '''
        return self.default is not inspect._empty

    @staticmethod
    def canwrap(param):
        ''' Returns True if Parameter can be used to wrap given function parameter.
            Returns False if param is keyword-only. '''
        return (
            param.kind not in (param.VAR_KEYWORD, param.KEYWORD_ONLY) and
            param.name not in ('self', 'cls')
        )


class ArgumentError(Exception):
    ''' This error is raised when stack argument parsing fails. '''
    pass
class Parser:
    ''' Base implementation of argument parsers
        that are used to parse arguments to stack commands.
    '''

    # Output size of this parser
    size = 1

    def __init__(self, parsefun=None):
        self.parsefun = parsefun

    def parse(self, argstring):
        ''' Parse the next argument from argstring. '''
        curarg, argstring = re_getarg.match(argstring).groups()
        return self.parsefun(curarg), argstring


class StringArg(Parser):
    ''' Argument parser that simply consumes the entire remaining text string. '''
    def parse(self, argstring):
        return argstring, ''


class AcidArg(Parser):
    ''' Argument parser for aircraft callsigns and group ids. '''
    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        acid = arg.upper()
        if acid in bs.traf.groups:
            idx = bs.traf.groups.listgroup(acid)
        else:
            idx = bs.traf.id2idx(acid)
            if idx < 0:
                raise ArgumentError(f'Aircraft with callsign {acid} not found')

            # Update ref position for navdb lookup
            bs.ref.lat = bs.traf.lat[idx]
            bs.ref.lon = bs.traf.lon[idx]
            bs.ref.acidx = idx
        return idx, argstring


class WpinrouteArg(Parser):
    ''' Argument parser for waypoints in an aircraft route. '''
    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        wpname = arg.upper()
        if bs.ref.acidx >= 0 and wpname in bs.traf.ap.route[bs.ref.acidx].wpname or wpname == '*':
            return wpname, argstring
        raise ArgumentError(f'{wpname} not found in the route of {bs.traf.id[bs.ref.acidx]}')

class WptArg(Parser):
    ''' Argument parser for waypoints.
        Makes 1 or 2 argument(s) into 1 position text to be used as waypoint

        Examples valid position texts:
        lat/lon : "N52.12,E004.23","N52'14'12',E004'23'10"
        navaid/fix: "SPY","OA","SUGOL"
        airport:   "EHAM"
        runway:    "EHAM/RW06" "LFPG/RWY23"
        Default values
    '''
    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        name = arg.upper()

        # Try aircraft first: translate a/c id into a valid position text with a lat,lon
        idx = bs.traf.id2idx(name)
        if idx >= 0:
            name = f'{bs.traf.lat[idx]},{bs.traf.lon[idx]}'

        # Check if lat/lon combination
        elif islat(name):
            # lat,lon ? Combine into one string with a comma
            arg, argstring = re_getarg.match(argstring).groups()
            name = name + "," + arg

        # apt,runway ? Combine into one string with a slash as separator
        elif argstring[:2].upper() == "RW" and name in bs.navdb.aptid:
            arg, argstring = re_getarg.match(argstring).groups()
            name = name + "/" + arg.upper()

        return name, argstring


class PosArg(Parser):
    ''' Argument parser for lat/lon positions.
        Makes 1 or 2 argument(s) into a lat/lon coordinate

        Examples valid position texts:
        lat/lon : "N52.12,E004.23","N52'14'12',E004'23'10"
        navaid/fix: "SPY","OA","SUGOL"
        airport:   "EHAM"
        runway:    "EHAM/RW06" "LFPG/RWY23"
        Default values
    '''
    # This parser's output size is 2 (lat, lon)
    size = 2

    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        argu = arg.upper()

        # Try aircraft first: translate a/c id into a valid position text with a lat,lon
        idx = bs.traf.id2idx(argu)
        if idx >= 0:
            return bs.traf.lat[idx], bs.traf.lon[idx], argstring

        # Check if lat/lon combination
        if islat(argu):
            nextarg, argstring = re_getarg.match(argstring).groups()
            bs.ref.lat = txt2lat(argu)
            bs.ref.lon = txt2lon(nextarg)
            return txt2lat(argu), txt2lon(nextarg), argstring

        # apt,runway ? Combine into one string with a slash as separator
        if argstring[:2].upper() == "RW" and argu in bs.navdb.aptid:
            arg, argstring = re_getarg.match(argstring).groups()
            argu = argu + "/" + arg.upper()

        posobj = Position(argu, bs.ref.lat, bs.ref.lon)
        if posobj.error:
            raise ArgumentError(f'{argu} is not a valid waypoint, airport, runway, or aircraft id.')

        # Update reference lat/lon
        bs.ref.lat = posobj.lat
        bs.ref.lon = posobj.lon
        bs.ref.hdg = posobj.refhdg

        return posobj.lat, posobj.lon, argstring


class PandirArg(Parser):
    ''' Parse pan direction commands. '''
    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        
        lat, lon = bs.scr.getviewctr()
        panmap = {'LEFT': (0.0, -0.5), 'RIGHT': (0.0, 0.5), 'UP': (0.5, 0.0), 'DOWN': (-0.5, 0.0)}
        delta = panmap.get(arg.upper(), None)
        if delta is None:
            raise ArgumentError(f'{arg} is not a valid pan direction')

        return lat + delta[0], lon + delta[1], argstring


class ColorArg(Parser):
    ''' Parse colour argument. '''
    def parse(self, argstring):
        arg, argstring = re_getarg.match(argstring).groups()
        try:
            if arg.isnumeric():
                g, argstring = re_getarg.match(argstring).groups()
                b, argstring = re_getarg.match(argstring).groups()
                return int(arg), int(g), int(b), argstring
            r, g, b = [int(255 * i) for i in colors.to_rgb(arg.upper())]
            return r, g, b, argstring
        except ValueError:
            raise ArgumentError(f'Could not parse "{arg}" as colour')


argparsers = {
    '*': None,
    'txt': Parser(str.upper),
    'word': Parser(str),
    'string': StringArg(),
    'float': Parser(float),
    'int': Parser(int),
    'onoff': Parser(txt2bool),
    'bool': Parser(txt2bool),
    'acid': AcidArg(),
    'wpinroute': WpinrouteArg(),
    'wpt': WptArg(),
    'latlon': PosArg(),
    'lat': PosArg(),
    'lon': None,
    'pandir': PandirArg(),
    'spd': Parser(txt2spd),
    'vspd': Parser(txt2vs),
    'alt': Parser(txt2alt),
    'hdg': Parser(lambda txt: txt2hdg(txt, bs.ref.lat, bs.ref.lon)),
    'time': Parser(txt2tim),
    'colour': ColorArg(),
    'color': ColorArg(),
}
